from __future__ import annotations

import ctypes
import hashlib
import re
import subprocess
from enum import Enum
from pathlib import Path
from typing import Any

from tabulate import tabulate

import pwndbg.aglib
import pwndbg.aglib.file
import pwndbg.aglib.memory
import pwndbg.aglib.vmmap
import pwndbg.color.message as message
import pwndbg.dbg_mod
import pwndbg.lib.cache
import pwndbg.lib.tempfile
import pwndbg.libc
from pwndbg.color import colorize
from pwndbg.color import generateColorFunction

ADDRESS_WRITABLE_PATTERN = re.compile(r"address(?:es)? (.*) (?:is|are) writable")
WRITABLE_COLON_PATTERN = re.compile(r"writable: (.*)")
EQUAL_NULL_PATTERN = re.compile(r"(.*) == NULL")
VALID_POSIX_SPAWN_FILE_ACTIONS_PATTERN = re.compile(r"(.*) <= 0")
VALID_ARGV_PATTERN = re.compile(r"(.+) is a valid argv")
VALID_ENVP_PATTERN = re.compile(r"(.+) is a valid envp")
IS_ALIGNED_PATTERN = re.compile(r"(.+) & 0xf == (\d+)")
IS_GOT_ADDRESS_PATTERN = re.compile(r"(.+) is the GOT address of libc")
CAST_PATTERN = re.compile(r"^\([s|u]\d+\)")
XMM_SHIFT = " >> "
CONSTRAINT_SEPARATOR = " || "
CAST_DEREF_MAPPING = {
    "(u16)": pwndbg.aglib.memory.u16,
    "(s16)": pwndbg.aglib.memory.s16,
    "(u32)": pwndbg.aglib.memory.u32,
    "(s32)": pwndbg.aglib.memory.s32,
    "(u64)": pwndbg.aglib.memory.u64,
    "(s64)": pwndbg.aglib.memory.s64,
}
CAST_MAPPING = {
    "(u16)": lambda x: ctypes.c_uint16(x).value,
    "(s16)": lambda x: ctypes.c_int16(x).value,
    "(u32)": lambda x: ctypes.c_uint32(x).value,
    "(s32)": lambda x: ctypes.c_int32(x).value,
    "(u64)": lambda x: ctypes.c_uint64(x).value,
    "(s64)": lambda x: ctypes.c_int64(x).value,
}
ONEGADGET_COLOR = {
    "light_green": lambda x: colorize(x, "\x1b[38;5;82m"),
    "light_purple": lambda x: colorize(x, "\x1b[38;5;153m"),
}
ONEGADGET_CACHEDIR: Path = pwndbg.lib.tempfile.cachedir("onegadget")


class CheckSatResult(Enum):
    SAT = 1
    UNSAT = 0
    UNKNOWN = -1

    def __str__(self) -> str:
        return self.name

    def __and__(self, other: CheckSatResult) -> CheckSatResult:
        if self == CheckSatResult.UNSAT or other == CheckSatResult.UNSAT:
            return CheckSatResult.UNSAT
        if self == CheckSatResult.UNKNOWN or other == CheckSatResult.UNKNOWN:
            return CheckSatResult.UNKNOWN
        return CheckSatResult.SAT

    def __or__(self, other: CheckSatResult) -> CheckSatResult:
        if self == CheckSatResult.SAT or other == CheckSatResult.SAT:
            return CheckSatResult.SAT
        if self == CheckSatResult.UNKNOWN or other == CheckSatResult.UNKNOWN:
            return CheckSatResult.UNKNOWN
        return CheckSatResult.UNSAT


SAT = CheckSatResult.SAT
UNSAT = CheckSatResult.UNSAT
UNKNOWN = CheckSatResult.UNKNOWN


class Lambda:
    """
    Modified from onegadget's Lambda class

    https://github.com/david942j/one_gadget/blob/65ce1dade70bf89e7496346ccf452ce5b2d139b3/lib/one_gadget/emulators/lambda.rb#L13
    """

    def __init__(self, obj: str | Lambda) -> None:
        self.immi = 0
        self.obj = obj
        self.deref_count = 0

    def __add__(self, other: int) -> Lambda:
        if not isinstance(other, int):
            raise ValueError(f"Expect other({other}) to be numeric.")

        if self.deref_count > 0:
            ret = Lambda(self)
        else:
            ret = Lambda(self.obj)
            ret.immi = self.immi
        ret.immi += other
        return ret

    def __sub__(self, other: int) -> Lambda:
        return self + (-other)

    def __str__(self) -> str:
        str_repr = ""
        str_repr += "[" * self.deref_count
        str_repr += str(self.obj) if self.obj is not None else ""
        str_repr += f"{self.immi:+#x}" if self.immi != 0 else ""
        str_repr += "]" * self.deref_count
        return str_repr

    def __repr__(self) -> str:
        return f"<Lambda obj={self.obj}, immi={self.immi}, deref_count={self.deref_count}>"

    @property
    def gdb_expr(self) -> str:
        # TODO: Don't use gdb.parse_and_eval here, directly fetching the value with `pwndbg.aglib.memory` would be better(?)
        obj = self.obj
        if isinstance(obj, str):
            if obj.startswith("xmm"):
                # Currently, "xmm\d >> {ptrsize*8}" only happens here:
                # https://github.com/david942j/one_gadget/blob/65ce1dade70bf89e7496346ccf452ce5b2d139b3/lib/one_gadget/emulators/x86.rb#L242-L248
                # So we can hardcode the shifting :p
                # TODO: Handle xmm register in a better way

                # TODO: In LLDB this syntax is invalid:
                # error: <user expression 62>:1:23: illegal vector component name 'v'
                #     1 | ((unsigned long)($xmm0.v2_int64[1]))
                #       |                       ^~~~~~~~~
                #  while parsing (u64)xmm0 >> 64 for argv[1]

                bits = pwndbg.aglib.arch.ptrbits
                if XMM_SHIFT in obj:
                    obj = obj.replace(XMM_SHIFT + str(bits), f".v{128 // bits}_int{bits}[1]")
                else:
                    obj += f".v{128 // bits}_int{bits}[0]"
            obj = f"(unsigned long)(${obj})"
        elif isinstance(obj, Lambda):
            obj = obj.gdb_expr
        else:
            raise ValueError(f"Unsupported obj: {obj}")
        str_repr = ""
        if self.deref_count > 0:
            str_repr += f"{'*' * self.deref_count}(unsigned long{'*' * self.deref_count})"
        str_repr += "("
        str_repr += obj
        str_repr += f"{self.immi:+#x}" if self.immi != 0 else ""
        str_repr += ")"
        return str_repr

    @property
    def color_str(self) -> str:
        str_repr = ""
        str_repr += "[" * self.deref_count
        if isinstance(self.obj, Lambda):
            str_repr += str(self.obj)
        elif self.obj:
            str_repr += colorize_reg(str(self.obj))
        if self.immi != 0:
            str_repr += "-" if self.immi < 0 else "+"
            str_repr += colorize_integer(hex(abs(self.immi)))
        str_repr += "]" * self.deref_count
        return str_repr

    def deref(self) -> Lambda:
        ret = Lambda(self.obj)
        ret.immi = self.immi
        ret.deref_count = self.deref_count + 1
        return ret

    def deref_(self) -> None:
        self.deref_count += 1

    def ref(self) -> Lambda:
        if self.deref_count <= 0:
            raise ValueError("Cannot reference anymore!")
        self.deref_count -= 1
        return self

    def evaluate(self, context: dict[Any, Any]) -> int | Lambda:
        if self.deref_count > 0 or (self.obj and self.obj not in context):
            raise ValueError(f"Can't eval {self}")
        return context[self.obj] + self.immi

    @staticmethod
    def parse(argument: str, predefined: dict[Any, Any] = {}) -> int | Lambda:
        if not argument or argument == "!":
            return 0
        try:
            return int(argument, 0)
        except ValueError:
            pass

        # nested []
        if argument.startswith("["):
            ridx = argument.rindex("]")
            immi = Lambda.parse(argument[(ridx + 1) :])
            lm = Lambda.parse(argument[1:ridx], predefined)
            if not isinstance(lm, Lambda) or not isinstance(immi, int):
                raise ValueError(f"Unsupported instruction argument: {argument}")
            lm = lm.deref()
            if immi:
                lm += immi
            return lm

        base, disp = Lambda.mem_obj(argument)
        obj = predefined.get(base, Lambda(base))
        if disp:
            obj += disp
        return obj

    @staticmethod
    def mem_obj(arg: str) -> tuple[str, int]:
        tokens = re.split(r"[+\-]", arg)
        if len(tokens) == 1:
            return tokens[0], 0
        if len(tokens) == 2:
            return tokens[0], int(tokens[1], 0) * (-1 if "-" in arg else 1)
        raise ValueError(f"Unsupported instruction argument: {arg}")


def colorize_reg(x: object) -> str:
    return generateColorFunction("light_green", ONEGADGET_COLOR)(x)


def colorize_integer(x: object) -> str:
    return generateColorFunction("light_purple", ONEGADGET_COLOR)(x)


def colorize_psuedo_code(code: str) -> str:
    """
    Colorize the pseudo code of onegadget
    """
    args_start = code.find("(")
    output = code[:args_start]
    args = []
    for arg in code[args_start + 1 : -1].split(", "):
        if arg[0] != '"' and arg != "environ":
            lambda_expr = Lambda.parse(arg)
            if isinstance(lambda_expr, Lambda):
                args.append(lambda_expr.color_str)
            else:
                args.append(colorize_integer(lambda_expr))
        else:
            args.append(arg)
    output += "(" + ", ".join(args) + ")"
    return output


def compute_file_hash(filename: str) -> str:
    """
    Compute the MD5 hash of the file, return the hash
    """
    h = hashlib.md5()
    with open(filename, "rb") as f:
        h.update(f.read())
    return h.hexdigest()


@pwndbg.lib.cache.cache_until("start", "objfile")
def run_onegadget() -> str:
    """
    Run onegadget and return the output
    """
    libc_path = pwndbg.aglib.file.get_file(str(pwndbg.libc.filepath()))
    # We need cache because onegadget might be slow
    cache_file: Path = ONEGADGET_CACHEDIR / compute_file_hash(libc_path)
    if cache_file.exists():
        # Cache hit
        with open(cache_file) as f:
            return f.read()
    # Cache miss
    output = subprocess.check_output(["one_gadget", "--level=100", libc_path], text=True)
    with open(cache_file, "w") as f:
        f.write(output)
    return output


def parse_expression(expr: str) -> tuple[int | None, str, str | None]:
    """
    Parse expression, return the result, colorized string and error message
    """
    # Remove cast
    match = CAST_PATTERN.match(expr)
    cast = match.group(0) if match else ""
    if cast:
        expr = expr[len(cast) :]
    if expr.startswith("("):
        # Remove the first and last parenthesis
        expr = expr[1:-1]
    lambda_expr = Lambda.parse(expr)
    if not isinstance(lambda_expr, Lambda):
        return lambda_expr, expr, None

    gdb_expr = lambda_expr.gdb_expr
    try:
        if cast:
            if gdb_expr.startswith("*"):
                # Remove the first *, we use cast to handle it instead of unsigned long
                gdb_expr = gdb_expr.lstrip("*")
                # Now gdb_expr is a pointer, we need dereference it with cast
                addr = int(pwndbg.dbg.selected_inferior().evaluate_expression(gdb_expr))
                result = CAST_DEREF_MAPPING[cast](addr)
            else:
                addr = int(pwndbg.dbg.selected_inferior().evaluate_expression(gdb_expr))
                result = CAST_MAPPING[cast](addr)
        else:
            addr = int(pwndbg.dbg.selected_inferior().evaluate_expression(gdb_expr))
            result = addr

        return result, f"{cast}{lambda_expr.color_str}", None
    except pwndbg.dbg_mod.Error as e:
        return None, f"{cast}{lambda_expr.color_str}", str(e)


def check_stack_argv(expr: str) -> tuple[CheckSatResult, str]:
    """
    Check argv that's on the stack, return the result and the message
    """
    output_msg = ""
    exprs = expr[1:-1].split(", ")

    n = 0
    for expr in exprs:
        if "sh" in expr or "-c" in expr:
            output_msg += f"argv[{n}] = {expr}\n"
            n += 1
            continue

        if expr == "NULL":
            output_msg += f"argv[{n}] = {expr}\n"
            return UNKNOWN, output_msg

        if expr == "...":
            output_msg += f"argv doesn't end, please check argv[{n}..n] manually\n"
            return UNKNOWN, output_msg

        result, color_str, err = parse_expression(expr)
        if err is not None:
            output_msg += f"{err} while parsing {color_str} for argv[{n}]\n"
            return UNSAT, output_msg
        if result == 0:
            if n > 1 and "-c" in exprs[n - 1]:
                output_msg += f'argv[{n}] = {color_str} = NULL, {color_str} can\'t be NULL because argv[{n - 1}] = "-c"\n'
                return UNSAT, output_msg
            output_msg += f"argv[{n}] = {color_str} = NULL\n"
            if n > 1:
                return UNKNOWN, output_msg
            return SAT, output_msg
        page = pwndbg.aglib.vmmap.find(result)
        if page is None or not page.read:
            output_msg += (
                f"argv[{n}] = {color_str} = {result:#x}, {color_str} is not a valid address\n"
            )
            return UNSAT, output_msg
        if n > 0:
            output_msg += f"argv[{n}] = {color_str} = {result:#x} -> {bytes(pwndbg.aglib.memory.string(result))!r}\n"
        else:
            output_msg += (
                f"argv[{n}] = {color_str} = {result:#x}, {color_str} is a readable address\n"
            )
        n += 1

    return SAT, output_msg


def check_non_stack_argv(expr: str) -> tuple[CheckSatResult, str]:
    """
    Check argv that's not on the stack, return the result and the message
    """
    output_msg = ""
    argv, color_str, err = parse_expression(expr)
    if err is not None:
        # We don't have to print the error message here, it should be printed already
        return UNSAT, f"{err} while parsing {color_str}\n"

    output_msg += f"Assume argv = {color_str} = {argv:#x}, checking the content of argv\n"

    n = 0
    while True:
        try:
            argv_n = pwndbg.aglib.memory.read_pointer_width(argv + n * pwndbg.aglib.arch.ptrsize)
        except pwndbg.dbg_mod.Error:
            output_msg += f"&argv[{n}] = {argv + n * pwndbg.aglib.arch.ptrsize:#x}, {argv + n * pwndbg.aglib.arch.ptrsize:#x} is a invalid address\n"
            return UNSAT, output_msg
        if argv_n == 0:
            if n > 1:
                output_msg += f"argv[{n}] is NULL, {color_str} might be a valid argv\n"
                return UNKNOWN, output_msg
            # {whatever_but_readable, NULL} is always a valid argv
            output_msg += f"argv[{n}] is NULL, {color_str} is a valid argv\n"
            return SAT, output_msg
        page = pwndbg.aglib.vmmap.find(argv_n)
        if page is None or not page.read:
            output_msg += f"argv[{n}] = {argv_n:#x}, {argv_n:#x} is a invalid address\n"
            return UNSAT, output_msg
        output_msg += f"argv[{n}] = {argv_n:#x} -> {bytes(pwndbg.aglib.memory.string(argv_n))!r}\n"
        n += 1


def check_argv(expr: str) -> tuple[CheckSatResult, str]:
    """
    Check argv, return the result and the message
    """
    if expr.startswith("{"):
        return check_stack_argv(expr)
    return check_non_stack_argv(expr)


def check_envp(expr: str) -> tuple[bool, str]:
    """
    Check envp, return the result and the message
    """
    output_msg = ""
    if expr.startswith("{"):
        # Note: we don't have to handle this case for now, but might need to implement it in the future
        return False, output_msg

    envp, color_str, err = parse_expression(expr)
    if err is not None:
        # we don't have to print the error message here, it should be printed already
        return False, f"{err} while parsing {color_str}\n"

    output_msg += f"Assume envp = {color_str} = {envp:#x}, checking the content of envp\n"

    # we need to make sure envp[0] is a valid pointer
    # until envp[n] is NULL
    n = 0
    while True:
        try:
            envp_n = pwndbg.aglib.memory.read_pointer_width(envp + n * pwndbg.aglib.arch.ptrsize)
        except pwndbg.dbg_mod.Error:
            output_msg += f"&envp[{n}] = {envp + n * pwndbg.aglib.arch.ptrsize:#x}, {envp + n * pwndbg.aglib.arch.ptrsize:#x} is a invalid address\n"
            return False, output_msg
        if envp_n == 0:
            output_msg += f"envp[{n}] is NULL, {color_str} is a valid envp\n"
            return True, output_msg
        page = pwndbg.aglib.vmmap.find(envp_n)
        if page is None or not page.read:
            output_msg += f"envp[{n}] = {envp_n:#x}, {envp_n:#x} is a invalid address\n"
            return False, output_msg
        output_msg += f"envp[{n}] = {envp_n:#x}, {envp_n:#x} is a readable address\n"
        n += 1


def check_constraint(constraint: str) -> tuple[CheckSatResult, str]:
    """
    Parse constraint, return the result and the message
    """
    output_msg = ""
    if CONSTRAINT_SEPARATOR in constraint:
        final_result = UNSAT
        for sub_constraint in constraint.split(CONSTRAINT_SEPARATOR):
            result1, msg = check_constraint(sub_constraint)
            output_msg += msg
            final_result = final_result | result1
            if final_result == SAT:
                return SAT, output_msg
        return final_result, output_msg

    passed = False

    # https://github.com/david942j/one_gadget/blob/65ce1dade70bf89e7496346ccf452ce5b2d139b3/lib/one_gadget/gadget.rb#L63-L88
    if EQUAL_NULL_PATTERN.match(constraint):
        expr = EQUAL_NULL_PATTERN.match(constraint).group(1)
        result, color_str, err = parse_expression(expr)
        if err is None:
            passed = result == 0
            output_msg += (
                f"{color_str} = {result:#x}, {color_str} {'==' if passed else '!='} NULL\n"
            )
        else:
            output_msg += f"{err} while parsing {color_str}\n"
    elif ADDRESS_WRITABLE_PATTERN.match(constraint):
        exprs = ADDRESS_WRITABLE_PATTERN.match(constraint).group(1)
        for expr in exprs.split(", "):
            result, color_str, err = parse_expression(expr)
            if err is None:
                page = pwndbg.aglib.vmmap.find(result)
                passed = page is not None and page.write
                output_msg += f"{color_str} = {result:#x}, {color_str} is {'' if passed else 'not '}writable\n"
            else:
                output_msg += f"{err} while parsing {color_str}\n"
                passed = False
            if not passed:
                break
    elif WRITABLE_COLON_PATTERN.match(constraint):
        expr = WRITABLE_COLON_PATTERN.match(constraint).group(1)
        result, color_str, err = parse_expression(expr)
        if err is None:
            page = pwndbg.aglib.vmmap.find(result)
            passed = page is not None and page.write
            output_msg += (
                f"{color_str} = {result:#x}, {color_str} is {'' if passed else 'not '}writable\n"
            )
        else:
            output_msg += f"{err} while parsing {color_str}\n"
    elif VALID_ARGV_PATTERN.match(constraint):
        arr = VALID_ARGV_PATTERN.match(constraint).group(1)
        passed, msg = check_argv(arr)
        output_msg += msg
    elif VALID_ENVP_PATTERN.match(constraint):
        expr = VALID_ENVP_PATTERN.match(constraint).group(1)
        passed, msg = check_envp(expr)
        output_msg += msg
    elif VALID_POSIX_SPAWN_FILE_ACTIONS_PATTERN.match(constraint):
        expr = VALID_POSIX_SPAWN_FILE_ACTIONS_PATTERN.match(constraint).group(1)
        result, color_str, err = parse_expression(expr)
        if err is None:
            assert isinstance(result, int)  # somehow mypy is complaining without this :/
            passed = result <= 0
            output_msg += f"{color_str} = {result:#x}, {color_str} {'<=' if passed else '>'} 0\n"
        else:
            output_msg += f"{err} while parsing {color_str}\n"
    elif IS_ALIGNED_PATTERN.match(constraint):
        expr, value = IS_ALIGNED_PATTERN.match(constraint).groups()
        value = int(value, 0)
        result, color_str, err = parse_expression(expr)
        if err is None:
            assert isinstance(result, int)  # somehow mypy is complaining without this :/
            passed = result & 0xF == value
            output_msg += f"{color_str} = {result:#x}, {color_str} & 0xf {'==' if passed else '!='} {value:#x}\n"
        else:
            output_msg += f"{err} while parsing {color_str}\n"
    elif IS_GOT_ADDRESS_PATTERN.match(constraint):
        expr = IS_GOT_ADDRESS_PATTERN.match(constraint).group(1)
        result, color_str, err = parse_expression(expr)
        if err is None:
            got_plt_address = pwndbg.libc.section_address_by_name(".got.plt")
            passed = result == got_plt_address
            output_msg += f"{color_str} = {result:#x}, {color_str} is {'' if passed else 'not '}the GOT address ({got_plt_address:#x}) of libc\n"
        else:
            output_msg += f"{err} while parsing {color_str}\n"
    else:
        raise ValueError(f"Unsupported constraint: {constraint}")

    return CheckSatResult(passed), output_msg


def check_gadget(
    gadget: str, show_unsat: bool = False, no_unknown: bool = False, verbose: bool = False
) -> CheckSatResult:
    """
    Check status of each gadget, return the gadget's status
    """
    lines = gadget.splitlines()
    # First line is offset of the gadget and C pseudo code
    offset, pseudo_code = lines[0].split(maxsplit=1)
    offset = int(offset, 16)
    output_msg = colorize_integer(hex(offset)) + " " + colorize_psuedo_code(pseudo_code) + "\n"
    verbose_msg = ""
    # From third line, there're the constraints
    is_valid_gadget = SAT
    result_list = []
    for line in lines[2:]:
        line = line.strip()
        result, msg = check_constraint(line)
        is_valid_gadget = is_valid_gadget & result
        if result == SAT:
            if verbose:
                verbose_msg += msg + message.success(f"SAT: {line}") + "\n"
            result_list.append((message.success(result), message.success(line)))
        elif result == UNSAT:
            if verbose:
                verbose_msg += msg + message.error(f"UNSAT: {line}") + "\n"
            result_list.append((message.error(result), message.error(line)))
        else:
            if verbose:
                verbose_msg += msg + message.warn(f"UNKNOWN: {line}") + "\n"
            result_list.append((message.warn(result), message.warn(line)))

    if verbose:
        output_msg += verbose_msg
    output_msg += tabulate(result_list, headers=["Result", "Constraint"], tablefmt="grid") + "\n"

    if is_valid_gadget == SAT:
        print(output_msg)
    elif is_valid_gadget == UNSAT and show_unsat:
        print(output_msg)
    elif is_valid_gadget == UNKNOWN and not no_unknown:
        print(output_msg)

    return is_valid_gadget


def find_gadgets(
    show_unsat: bool = False, no_unknown: bool = False, verbose: bool = False
) -> dict[CheckSatResult, int]:
    """
    Find gadgets by parsing the output of onegadget, return there's any valid gadget
    """
    gadgets = run_onegadget().split("\n\n")
    gadgets_count = {SAT: 0, UNSAT: 0, UNKNOWN: 0}
    for gadget in gadgets:
        result = check_gadget(gadget, show_unsat=show_unsat, no_unknown=no_unknown, verbose=verbose)
        gadgets_count[result] += 1

    return gadgets_count



from __future__ import annotations

import argparse

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.stack
import pwndbg.aglib.typeinfo
import pwndbg.aglib.vmmap
import pwndbg.color.memory as mem_color
import pwndbg.commands
from pwndbg.commands import CommandCategory
from pwndbg.lib.memory import Page

parser = argparse.ArgumentParser(
    description="Shows offsets of the specified address from various useful locations."
)
parser.add_argument("address", nargs="?", default="$pc", help="Address to inspect", type=int)


def print_line(name, addr, first, second, op, width=20) -> None:
    print(
        f"{name.rjust(width)} {mem_color.get(addr)} = {mem_color.get(first) if not isinstance(first, str) else first.ljust(len(hex(addr).rstrip('L')))} {op} {second:#x}"
    )


def xinfo_stack(page: Page, addr: int) -> None:
    # If it's a stack address, print offsets to top and bottom of stack, as
    # well as offsets to current stack and base pointer (if used by debuggee)

    sp = pwndbg.aglib.regs.sp
    frame = pwndbg.aglib.regs.read_reg(pwndbg.aglib.regs.frame)
    frame_mapping = pwndbg.aglib.vmmap.find(frame)

    print_line("Stack Top", addr, page.vaddr, addr - page.vaddr, "+")
    print_line("Stack End", addr, page.end, page.end - addr, "-")
    print_line("Stack Pointer", addr, sp, addr - sp, "+")

    if frame_mapping and page.vaddr == frame_mapping.vaddr:
        print_line("Frame Pointer", addr, frame, frame - addr, "-")

    canary_value = pwndbg.commands.canary.canary_value()[0]

    if canary_value is not None:
        all_canaries = list(
            pwndbg.search.search(
                pwndbg.aglib.arch.pack(canary_value), mappings=pwndbg.aglib.stack.get().values()
            )
        )
        follow_canaries = sorted(filter(lambda a: a > addr, all_canaries))
        if follow_canaries is not None and len(follow_canaries) > 0:
            nxt = follow_canaries[0]
            print_line("Next Stack Canary", addr, nxt, nxt - addr, "-")


def xinfo_mmap_file(page: Page, addr: int) -> None:
    # If it's an address pointing into a memory mapped file, print offsets
    # to beginning of file in memory and on disk

    file_name = page.objfile

    region_start = pwndbg.aglib.vmmap.addr_region_start(addr)
    if region_start is None:
        print("The file is not contiguous in memory.")
        return

    # print offset from ELF base load address
    rva = addr - region_start
    print_line("File (Base)", addr, region_start, rva, "+")

    # find possible LOAD segments that designate memory and file backings
    containing_loads = [
        seg
        for seg in pwndbg.aglib.elf.get_containing_segments(file_name, region_start, addr)
        if seg["p_type"] == "PT_LOAD"
    ]

    for segment in containing_loads:
        if segment["p_type"] == "PT_LOAD" and addr < segment["x_vaddr_mem_end"]:
            offset = addr - segment["p_vaddr"]
            print_line("File (Segment)", addr, segment["p_vaddr"], offset, "+")
            break

    for segment in containing_loads:
        if segment["p_type"] == "PT_LOAD" and addr < segment["x_vaddr_file_end"]:
            file_offset = segment["p_offset"] + (addr - segment["p_vaddr"])
            print_line("File (Disk)", addr, file_name, file_offset, "+")
            break
    else:
        print(f"{'File (Disk)'.rjust(20)} {mem_color.get(addr)} = [not file backed]")

    containing_sections = pwndbg.aglib.elf.get_containing_sections(file_name, region_start, addr)
    if len(containing_sections) > 0:
        print("\n Containing ELF sections:")
        for sec in containing_sections:
            print_line(sec["x_name"], addr, sec["sh_addr"], addr - sec["sh_addr"], "+")


def xinfo_default(page: Page, addr: int) -> None:
    # Just print the distance to the beginning of the mapping
    print_line("Mapped Area", addr, page.vaddr, addr - page.vaddr, "+")


@pwndbg.commands.Command(parser, category=CommandCategory.MEMORY)
@pwndbg.commands.OnlyWhenRunning
def xinfo(address: int) -> None:
    address &= pwndbg.aglib.arch.ptrmask

    page = pwndbg.aglib.vmmap.find(address)

    if page is None:
        print(f"\n  Virtual address {address:#x} is not mapped.")
        return

    print(f"Extended information for virtual address {mem_color.get(address)}:")

    print("\n  Containing mapping:")
    print(mem_color.get(address, text=str(page)))

    print("\n  Offset information:")

    if page.is_stack:
        xinfo_stack(page, address)
    else:
        xinfo_default(page, address)

    if page.is_memory_mapped_file:
        xinfo_mmap_file(page, address)



from __future__ import annotations

import argparse
import math

import pwndbg.aglib
import pwndbg.aglib.kernel
import pwndbg.aglib.memory
import pwndbg.aglib.proc
import pwndbg.chain
import pwndbg.color as color
import pwndbg.color.context as ctx_color
import pwndbg.color.message as message
import pwndbg.commands
import pwndbg.commands.kcurrent
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(description="Performs pagewalk.")
parser.add_argument("vaddr", type=str, help="virtual address to walk")
parser.add_argument("--pgd", dest="entry", type=str, default=None, help="")

PAGETYPES = (
    "buddy",
    "offline",
    "table",
    "guard",
    "hugetlb",
    "slab",
    "zsmalloc",
    "unaccepted",
)


def page_type(page) -> str:
    names = PAGETYPES
    page_type_val = pwndbg.aglib.memory.s32(page + 0x30)
    if page_type_val == -1:
        return "initialized"
    if page_type_val >= 0:
        return f"mapcount: {page_type_val}"
    page_type_val = pwndbg.aglib.memory.u32(page + 0x30)
    krelease = pwndbg.aglib.kernel.krelease()
    if not krelease or krelease >= (6, 12):
        idx = (page_type_val >> 24) - 0xF0
        if idx < len(names):
            return names[idx]
    elif krelease >= (6, 11):
        names = names[:-1][::-1]
        for i in range(len(names)):
            if page_type_val & (1 << (i + 24)) == 0:
                return names[i]
    elif krelease >= (6, 10):
        names = names[:6]
        for i in range(len(names)):
            if page_type_val & (1 << (7 + i)) == 0:
                return names[i]
    elif krelease >= (5, 0):
        names = names[:5]
        for i in range(len(names)):
            if page_type_val & (1 << (7 + i)) == 0:
                return names[i]
    return "unknown"


def page_info(page) -> None:
    try:
        refcount = pwndbg.aglib.memory.u32(page + 0x34)
        print(
            f"{color.green('page')} @ {color.yellow(hex(page))} [{page_type(page)}, refcount: {refcount}]"
        )
    except (ValueError, TypeError):
        print(message.warn("invalid page address"))


@pwndbg.commands.Command(parser, category=CommandCategory.KERNEL)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWhenPagingEnabled
@pwndbg.aglib.proc.OnlyWithArch(["x86-64", "aarch64"])
def pagewalk(vaddr, entry=None) -> None:
    if entry is not None:
        entry = int(pwndbg.dbg.selected_frame().evaluate_expression(entry))
    if pwndbg.aglib.memory.is_kernel(entry):
        entry = pwndbg.aglib.kernel.pagewalk(entry, virt=False).phys
    vaddr = int(pwndbg.dbg.selected_frame().evaluate_expression(vaddr))
    result = pwndbg.aglib.kernel.pagewalk(vaddr, entry)
    for level in result.levels[::-1]:
        name, entry, vaddr, idx = level.name, level.entry, level.virt, level.idx
        if name is None or entry is None or vaddr is None or idx is None:
            break
        pageflags = pwndbg.aglib.kernel.bitflags(level)
        flags = ""
        arrow_right = pwndbg.chain.c.arrow(f"{pwndbg.chain.config_arrow_right}")
        if pwndbg.aglib.arch.name == "x86-64":
            name = name.ljust(3, " ")
        nbits = pwndbg.aglib.kernel.page_shift() - math.ceil(
            math.log2(pwndbg.aglib.arch.ptrsize)
        )  # each idx has that many bits
        idxlen = len(str((1 << nbits) - 1))
        if entry is not None:
            flags = f"[{idx:0{idxlen}}] {arrow_right} {name + 'e'}: {ctx_color.format_flags(entry, pageflags, entry)}"
        print(f"{color.blue(name)} @ {color.yellow(hex(vaddr))}{flags}")
    vaddr = result.virt
    if vaddr is None:
        print(message.warn("address is not mapped"))
        return
    phys = result.phys
    if phys:
        phys = color.yellow(hex(phys))
    print(f"pagewalk result: {color.green(hex(vaddr))} [phys: {phys}]")


def paging_print_helper(name, addr):
    if addr is None:
        return
    print(f"{color.green(name)}: {color.yellow(hex(addr))}")


p2v_parser = argparse.ArgumentParser(
    description="Translate physical address to its corresponding virtual address."
)
p2v_parser.add_argument("paddr", type=str, help="")


@pwndbg.commands.Command(p2v_parser, category=CommandCategory.KERNEL)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWithKernelSymbols
@pwndbg.commands.OnlyWhenPagingEnabled
@pwndbg.aglib.proc.OnlyWithArch(["x86-64", "aarch64"])
def p2v(paddr) -> None:
    paddr = int(pwndbg.dbg.selected_frame().evaluate_expression(paddr))
    try:
        vaddr = pwndbg.aglib.kernel.phys_to_virt(paddr)
        paging_print_helper("Virtual address", vaddr)
        page = pwndbg.aglib.kernel.virt_to_page(vaddr)
        page_info(page)
    except Exception:
        print(message.warn("physical to virtual address failed, invalid physical address?"))


v2p_parser = argparse.ArgumentParser(
    description="Translate virtual address to its corresponding physmap address."
)
v2p_parser.add_argument("vaddr", type=str, help="")


@pwndbg.commands.Command(v2p_parser, category=CommandCategory.KERNEL)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWithKernelSymbols
@pwndbg.commands.OnlyWhenPagingEnabled
@pwndbg.aglib.proc.OnlyWithArch(["x86-64", "aarch64"])
def v2p(vaddr) -> None:
    vaddr = int(pwndbg.dbg.selected_frame().evaluate_expression(vaddr))
    result = pwndbg.aglib.kernel.pagewalk(vaddr)  # more accurate
    entry, paddr = result.entry, result.phys
    if not entry or paddr is None:
        print(message.warn("virtual to physical address failed, unmapped virtual address?"))
        return
    paging_print_helper("Physical address", paddr)
    # paddr is the physmap address which is a virtual address
    page = pwndbg.aglib.kernel.phys_to_page(paddr)
    page_info(page)


page_parser = argparse.ArgumentParser(
    description="Convert a pointer to a `struct page` to its corresponding virtual address."
)
page_parser.add_argument("page", type=str, help="")


@pwndbg.commands.Command(page_parser, category=CommandCategory.KERNEL)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWithKernelSymbols
@pwndbg.commands.OnlyWhenPagingEnabled
@pwndbg.aglib.proc.OnlyWithArch(["x86-64", "aarch64"])
def pageinfo(page) -> None:
    page = int(pwndbg.dbg.selected_frame().evaluate_expression(page))
    try:
        vaddr = pwndbg.aglib.kernel.page_to_virt(page)
        paging_print_helper("Virtual address", vaddr)
        page_info(page)
    except Exception:
        print(message.warn("invalid page struct pointer"))



from __future__ import annotations

import hashlib
import logging
import os
import shutil
import subprocess
import sys
from pathlib import Path


def hash_file(file_path: str | Path) -> str:
    with open(file_path, "rb") as f:
        file_hash = hashlib.sha256()
        while True:
            chunk = f.read(8192)
            if not chunk:
                break
            file_hash.update(chunk)
    return file_hash.hexdigest()


def run_uv_install(
    binary_path: os.PathLike[str], src_root: Path, venv_path: Path, dev: bool = False
) -> tuple[str, str, int]:
    # Check if the package was installed using: `uv tool install --editable .[lldb,gdb]`
    # Tools are located at: ${HOME}/.local/share/uv/tools/${TOOL_NAME}/uv-receipt.toml
    is_tool_install = (venv_path / "uv-receipt.toml").exists()
    if is_tool_install:
        tool_name = venv_path.name
        command: list[str] = [str(binary_path), "tool", "upgrade", tool_name]
    else:
        # --inexact makes it so any installed extras aren't uninstalled
        # (it will also leave in dropped deps, but what can you do /shrug)
        command = [str(binary_path), "sync", "--inexact"]
        if dev:
            command.append("--all-groups")
    logging.debug(f"Updating deps with command: {' '.join(command)}")
    result = subprocess.run(command, capture_output=True, text=True, cwd=src_root)
    return result.stdout.strip(), result.stderr.strip(), result.returncode


def find_uv(venv_path: Path) -> Path | None:
    binary_path = shutil.which("uv", path=venv_path / "bin")
    if binary_path is not None:
        return Path(binary_path)

    return None


def is_dev_mode(venv_path: Path) -> bool:
    # If "dev.marker" exists in the venv directory, the user ran setup-dev.sh and is
    # considered a developer
    return (venv_path / "dev.marker").exists()


def update_deps(src_root: Path) -> None:
    venv_path = Path(sys.prefix)
    uv_lock_hash_path = venv_path / "uv.lock.hash"

    current_hash = hash_file(src_root / "uv.lock")
    logging.debug(f"Current uv.lock hash: {current_hash}")

    stored_hash = None
    if uv_lock_hash_path.exists():
        stored_hash = uv_lock_hash_path.read_text().strip()
        logging.debug(f"Stored uv.lock hash: {stored_hash}")
    else:
        logging.debug("No stored hash found")

    # If the hashes don't match, update the dependencies
    if current_hash == stored_hash:
        return

    print("Detected outdated Pwndbg dependencies (uv.lock). Updating.")
    uv_path = find_uv(venv_path)
    if uv_path is None:
        print(
            "'uv' was not found on the $PATH. Please ensure it is installed and on the path, "
            "or run `./setup.sh` to manually update Python dependencies."
        )
        return

    dev_mode = is_dev_mode(venv_path)
    stdout, stderr, return_code = run_uv_install(uv_path, src_root, venv_path, dev=dev_mode)
    if return_code == 0:
        uv_lock_hash_path.write_text(current_hash)

        # Only print the uv output if anything was actually updated
        if "No dependencies to install or update" not in stdout:
            print(stdout)
    else:
        print(stderr, file=sys.stderr)
        print("\x1b[31mERROR: Pwndbg failed to update with the above uv error.\x1b[0m")
        print(
            "\x1b[31m"
            "ERROR: Re-run with PWNDBG_NO_AUTOUPDATE=1 if you intend to run without an update."
            "\x1b[0m"
        )
        sys.exit(return_code)


def is_system_installation(src_root: Path) -> bool:
    # If pwndbg is installed in `/venv/lib/pythonX.Y/site-packages/pwndbg/`,
    # the `.pwndbg_root` file will not exist because `src_root` will point to the
    # `/venv/lib/pythonX.Y/site-packages/` directory, not the original source directory
    #
    # However, if pwndbg is installed in editable mode (our recommended way), this file will exist,
    # and the condition will be False, allowing auto-update.
    is_system_install = not (src_root / ".pwndbg_root").exists()
    if is_system_install:
        return True

    return False


def skip_autoupdate(src_root: Path) -> bool:
    no_auto_update = os.getenv("PWNDBG_NO_AUTOUPDATE") is not None
    if no_auto_update:
        return True

    if is_system_installation(src_root):
        return True

    return False


def verify_venv():
    src_root = Path(__file__).parent.parent.resolve()
    if skip_autoupdate(src_root):
        return

    update_deps(src_root)



"""
Add, load, show, edit, or delete custom structures.

For the compilation of the structures zig is being used under the hood, unless
`gcc-config-path` is specified.
"""

from __future__ import annotations

import os
import tempfile
from pathlib import Path

import pwndbg
import pwndbg.aglib.elf as elf
import pwndbg.lib.tempfile
from pwndbg.lib import Status

# Remeber loaded structures. This would be useful for 'remove-symbol-file'.
loaded_structures: dict[str, str] = {}

# Where generated structure source files are saved.
storage_location: Path = pwndbg.lib.tempfile.cachedir("custom-structures")


def compile_structure(struct_path: Path, compiled_path: str | None = None) -> tuple[str, Status]:
    """
    Compile a C file that contains custom struct definitions.

    Zig is being used unless `gcc-config-path` is specified. Naturally
    the compiled output file will have debug info.

    Arguments:
        struct_path: Path to the .c / .h file to compile.
        compiled_path: The path of the output file. If not specified, it will be
            an automatically generated filename in /tmp/.

    Returns:
        A (compiled_path, err) tuple.
    """
    if compiled_path is None:
        _, compiled_path = tempfile.mkstemp(prefix="custom-", suffix=".dbg")

    # -fno-eliminate-unused-debug-types is a handy gcc flag that lets us extract debug info from non-used defined structures.
    compiler_extra_flags = [
        str(struct_path),
        "-c",
        "-g",
        "-fno-eliminate-unused-debug-types",
        "-o",
        compiled_path,
    ]
    err: Status = elf.compile_with_flags(compiler_extra_flags)
    if err.is_failure():
        return "", err

    return compiled_path, Status()


def get_struct_path(name: str) -> Path:
    """
    Get a Path for a name (usually in ~/.cache/pwndbg/custom-structures/).
    """
    return storage_location / f"{name}.c"


def get_struct_path_if_exists(name: str) -> Path | None:
    """
    Get a Path for a name (usually in ~/.cache/pwndbg/custom-structures/) if the
    file exists, otherwise return None.
    """
    path: Path = get_struct_path(name)
    if path.exists():
        return path
    return None


def unload(name: str) -> None:
    """
    Unload structures from the debugger by set name.
    """
    struct_file: str | None = loaded_structures.pop(name, None)
    if struct_file is not None:
        pwndbg.dbg.selected_inferior().remove_symbol_file(struct_file)


def remove(name: str) -> Status:
    """
    Unload structures from the debugger and delete the backing file by set name.
    """
    struct_path: Path | None = get_struct_path_if_exists(name)
    if struct_path is None:
        return Status.fail("No custom structure was found with the given name!")

    unload(name)
    os.unlink(struct_path)
    return Status()


def load_with_path(name: str, struct_path: Path) -> Status:
    """
    Load structures from set name `name`, located at `struct_path` into the debbuger.

    Requires the set to have already been added.
    """
    unload(name)

    outfile, err = compile_structure(struct_path)
    if err.is_failure():
        return err

    pwndbg.dbg.selected_inferior().add_symbol_file(outfile)
    loaded_structures[name] = outfile
    os.unlink(outfile)
    return Status()


def load(name: str) -> Status:
    """
    Load structures from set name `name`.

    Requires the set to have already been added.
    """
    struct_path: Path | None = get_struct_path_if_exists(name)
    if struct_path is None:
        return Status.fail("No custom structure was found with the given name!")

    return load_with_path(name, struct_path)


def saved_names() -> list[str]:
    """
    Returns all set names.
    """
    res: list[str] = []
    for file in os.listdir(storage_location):
        if file.endswith(".c"):
            # Remove the ".c".
            name = os.path.splitext(file)[0]
            res.append(name)
    return res


def create_temp_header_file(content: str) -> Path:
    """
    Creates a temporary file with content `content` and returns the Path to it.
    """
    # We need to use `.c` rather than `.h` because `.h` causes zig/ld.lld to fail compilation:
    # "ld.lld: error: /home/user/.cache/zig/o/81a692b354ebabc64d08698dff966d67/page_structs.o: unknown file type"
    with tempfile.NamedTemporaryFile(delete=False, suffix=".c") as tmp_file:
        tmp_file.write(content.encode())
        return Path(tmp_file.name)


def add(name: str, content: str) -> Status:
    """
    Add structures defined in `content` (C code) by reference structure set name `name` and load
    them into the debugger.

    If a `name` structure set file already exists, it will be overwritten.

    The name will be prefixed with `_internal_` to separate from user-defined structures.
    """
    struct_path: Path = get_struct_path("_internal_" + name)
    # We don't care if it existed before.
    with open(struct_path, "w") as f:
        f.write(content)
    err = load_with_path(name, struct_path)
    return err



"""
The abstracted debugger interface.
"""

from __future__ import annotations

import contextlib
from collections.abc import Awaitable
from collections.abc import Callable
from collections.abc import Coroutine
from collections.abc import Generator
from collections.abc import Iterator
from collections.abc import Sequence
from enum import Enum
from typing import Any
from typing import Literal
from typing import TypedDict
from typing import TypeVar

import pwndbg.lib.cache
import pwndbg.lib.memory
from pwndbg.lib.arch import ArchDefinition
from pwndbg.lib.siginfo import SigInfo

dbg: Debugger = None

T = TypeVar("T")

# Increased by one on every stop event.
number_of_stops_since_birth: int = 0


@contextlib.contextmanager
def selection(target: T, get_current: Callable[[], T], select: Callable[[T], None]):
    """
    Debuggers have global state. Many of our queries require that we select a
    given object globally before we make them. When doing that, we must always
    be careful to return selection to its previous state before exiting. This
    class automatically manages the selection of a single object type.

    Upon entrace to the `with` block, the element given by `target` will be
    compared to the object returned by calling `get_current`. If they
    compare different, the value previously returned by `get_current` is
    saved, and the element given by `target` will be selected by passing it
    as an argument to `select`, and, after execution leaves the `with`
    block, the previously saved element will be selected in the same fashion
    as the first element.

    If the elements don't compare different, this is a no-op.
    """

    current = get_current()
    restore = False
    if current != target:
        select(target)
        restore = True

    try:
        yield
    finally:
        if restore:
            select(current)


class Error(Exception):
    pass


class NoInferior(Exception):
    def __init__(self) -> None:
        super().__init__("The debugger couldn't find a selected inferior.")


class DisassembledInstruction(TypedDict):
    addr: int
    asm: str
    length: int


class DebuggerType(Enum):
    GDB = 1
    LLDB = 2


class StopPoint:
    """
    The handle to either an insalled breakpoint or watchpoint.

    May be used in a `with` statement, in which case the stop point is
    automatically removed at the end of the statement. This allows for easy
    implementation of temporary breakpoints.
    """

    def remove(self) -> None:
        """
        Removes the breakpoint associated with this handle.
        """
        raise NotImplementedError()

    def set_enabled(self, enabled: bool) -> None:
        """
        Enables or disables this breakpoint.
        """
        raise NotImplementedError()

    def __enter__(self) -> StopPoint:
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """
        Automatic breakpoint removal.
        """
        self.remove()


class BreakpointLocation:
    """
    This is the location specification for a breakpoint.
    """

    address: int

    def __init__(self, address: int):
        self.address = address

    def __eq__(self, other: object) -> bool:
        if isinstance(other, BreakpointLocation):
            return self.address == other.address
        if isinstance(other, int):
            return self.address == other
        return False


class WatchpointLocation:
    """
    This is the location specification for a watchpoint.
    """

    address: int
    size: int
    watch_read: bool
    watch_write: bool

    def __init__(self, address: int, size: int, watch_read: bool, watch_write: bool):
        self.address = address
        self.size = size

        assert watch_read or watch_write, "Watchpoints must watch at least one of reads or writes"

        self.watch_read = watch_read
        self.watch_write = watch_write


class Registers:
    """
    A handle to the register values in a frame.
    """

    def by_name(self, name: str) -> Value | None:
        """
        Gets the value of a register if it exists, None otherwise.
        """
        raise NotImplementedError()


class SymbolLookupType(Enum):
    """
    Enum representing types of symbol lookups for filtering symbol searches.

    Attributes:
    - ANY: Represents searching for any symbol type (default).
    - FUNCTION: Represents searching specifically for function symbols.
    - VARIABLE: Represents searching specifically for variable symbols.
    """

    ANY = 1
    FUNCTION = 2
    VARIABLE = 3


class Frame:
    """
    A lightweight object referencing a stack frame in a given thread.

    It does not hold state, it allows us to ask the debugger for state
    (registers, symbols etc.) associated with this stack frame.

    This is a very short-lived object and easily becomes invalid so do not
    store it, and especially not between debugger stops.
    """

    def lookup_symbol(
        self,
        name: str,
        *,
        type: SymbolLookupType = SymbolLookupType.ANY,
    ) -> Value | None:
        """
        Looks up and returns the address of a symbol in current frame by its name.

        Parameters:
        - name (str): The name of the symbol to look up.
        - type (SymbolLookupType, optional): The type of symbol to search for. Defaults
          to SymbolLookupType.ANY.

        Returns:
        - pwndbg.dbg_mod.Value | None: The value of the symbol if found, or None if not found.

        Raises:
        - pwndbg.dbg_mod.Error: If symbol name contains invalid characters
        """
        raise NotImplementedError()

    def evaluate_expression(self, expression: str, lock_scheduler: bool = False) -> Value:
        """
        Evaluate the given expression in the context of this frame, and
        return a `Value`.

        # `lock_scheduler`
        Additionally, callers of this function might specify that they want to
        enable scheduler locking during the evaluation of this expression. This
        is a GDB-only option, and is intended for cases in which the result
        would be incorrect without it enabled, when running in GDB. Other
        debuggers should ignore this parameter.
        """
        raise NotImplementedError()

    def regs(self) -> Registers:
        """
        Access the values of the registers in this frame.
        """
        raise NotImplementedError()

    def reg_write(self, name: str, val: int) -> bool:
        """
        Sets the value of the register with the given name to the given value.
        Returns true if the register exists, false othewise. Throws an exception
        if the register exists but cannot be written to.
        """
        raise NotImplementedError()

    def pc(self) -> int:
        """
        The value of the program counter for this frame.
        """
        raise NotImplementedError()

    def sp(self) -> int:
        """
        The value of the stack pointer for this frame.
        """
        raise NotImplementedError()

    def start(self) -> int | None:
        """
        The start (highest) address of this frame. The return address
        is usually here.
        """
        raise NotImplementedError()

    def parent(self) -> Frame | None:
        """
        The parent frame of this frame, if it exists.
        """
        raise NotImplementedError()

    def child(self) -> Frame | None:
        """
        The child frame of this frame, if it exists.
        """
        raise NotImplementedError()

    def sal(self) -> tuple[str, int] | None:
        """
        The filename of the source code file associated with this frame, and the
        line number associated with it, if available.
        """
        raise NotImplementedError()

    @pwndbg.lib.cache.cache_until("forever")
    def stack_variables(self) -> tuple[tuple[int, int, str], ...]:
        """
        Get all stack variables (local variables and arguments) in current frame.

        Returns a tuple of (start_address, end_address, name) for each variable.
        Returns an empty tuple if no debug information is available or on error.

        Cached forever since a Frame is to be used ephemerally.
        """
        raise NotImplementedError()

    def __eq__(self, rhs: object) -> bool:
        """
        Whether this frame is the same as the given frame. Two frames are the
        same if they point to the same stack frame and have the same execution
        context.
        """
        raise NotImplementedError()

    def idx(self) -> int:
        """
        The index of this stack frame on the thread's stack. Index zero is the most
        fresh frame.
        """
        raise NotImplementedError()

    def __hash__(self) -> int:
        """
        The hash value of this stack frame. Needs to guarantee uniqueness both within
        a stop and *accross time* so caches can work properly.
        """
        # Looking at how the debugger implements the frame equality check can help you
        # figure this out.
        raise NotImplementedError()


class Thread:
    @contextlib.contextmanager
    def bottom_frame(self) -> Iterator[Frame]:
        """
        Frame at the bottom of the call stack for this thread.
        """
        raise NotImplementedError()

    def ptid(self) -> int | None:
        """
        The PTID of this thread, if available.
        """
        raise NotImplementedError()

    def index(self) -> int:
        """
        The unique index of this thread from the perspective of the debugger.
        """
        raise NotImplementedError()

    def siginfo(self) -> SigInfo | None:
        """
        The siginfo of this thread.
        """
        raise NotImplementedError()


class MemoryMap:
    """
    A wrapper around a sequence of memory ranges
    """

    pages: tuple[pwndbg.lib.memory.Page, ...]

    def __init__(self, pages: Sequence[pwndbg.lib.memory.Page]):
        self.pages = tuple(pages)

    def is_qemu(self) -> bool:
        """
        Returns whether this memory map was generated from a QEMU target.
        """
        raise NotImplementedError()

    def ranges(self) -> Sequence[pwndbg.lib.memory.Page]:
        """
        Returns all ranges in this memory map.
        """
        return self.pages

    def lookup_page(self, address: int) -> pwndbg.lib.memory.Page | None:
        # Binary search for the page
        lo = 0
        hi = len(self.pages) - 1
        while lo <= hi:
            mid = (hi + lo) // 2
            page = self.pages[mid]
            if page.start <= address:
                if address < page.end:
                    return page

                lo = mid + 1
            else:
                hi = mid - 1

        return None


class ExecutionController:
    def single_step(self) -> Awaitable[None]:
        """
        Steps to the next instruction.

        Throws `CancelledError` if a breakpoint or watchpoint is hit, the program
        exits, or if any other unexpected event that diverts execution happens
        while fulfulling the step.

        FIXME GDB:
        On GDB `stepi` will execute other threads. On LLDB not.
        Please use `set scheduler-locking step`
        """
        raise NotImplementedError()

    def cont(self, until: StopPoint) -> Awaitable[None]:
        """
        Continues execution until the given breakpoint or whatchpoint is hit.
        Continues execution on all threads.

        Throws `CancelledError` if a breakpoint or watchpoint is hit that is not
        the one given in `until`, the program exits, or if any other unexpected
        event happens.
        """
        raise NotImplementedError()

    def cont_selected_thread(self, until: StopPoint) -> Awaitable[None]:
        """
        Continues execution on single thread until the given breakpoint or whatchpoint is hit.
        Continues execution on selected thread.

        Throws `CancelledError` if a breakpoint or watchpoint is hit that is not
        the one given in `until`, the program exits, or if any other unexpected
        event happens.
        """
        raise NotImplementedError()


class Process:
    def threads(self) -> list[Thread]:
        """
        Returns a list containing the threads in this process.
        """
        raise NotImplementedError()

    def pid(self) -> int | None:
        """
        Returns the process ID of this process if it is alive.
        """
        raise NotImplementedError()

    def alive(self) -> bool:
        """
        Returns whether this process is alive.
        """
        raise NotImplementedError()

    def is_core_file(self) -> bool:
        """
        Returns whether this process is a coredump file.
        """
        raise NotImplementedError()

    def stopped_with_signal(self) -> bool:
        """
        Returns whether this process was stopped by a signal.
        """
        raise NotImplementedError()

    def stopped_at_breakpoint(self) -> bool:
        """
        Returns whether this process was stopped at a breakpoint.
        """
        raise NotImplementedError()

    def evaluate_expression(self, expression: str) -> Value:
        """
        Evaluate the given expression in the context of the current process, and
        return a `Value`.
        """
        raise NotImplementedError()

    def vmmap(self) -> MemoryMap:
        """
        Returns the virtual memory map of this process, as seen by the debugger.

        Generally, one should prefer `pwndbg.aglib.vmmap.get()` over this
        function, as this passes the raw information from the debugger more or
        less straight through, without applying more general Pwndbg enhancements
        to the memory map. This is the lower-level functionality on top of which
        the function in `aglib` is implemented.
        """
        raise NotImplementedError()

    def read_memory(self, address: int, size: int, partial: bool = False) -> bytearray:
        """
        Reads the requested number of bytes from the address given in the memory
        space of this process. Will read as many bytes as possible starting at
        that location, and returns how many were read.

        Throws an exception if reading fails and partial is False.
        """
        raise NotImplementedError()

    def write_memory(self, address: int, data: bytearray, partial: bool = False) -> int:
        """
        Writes as many bytes from the given data buffer as possible into the
        given address in the memory space of this process.

        Throws an exception if writing fails and partial is False.
        """
        raise NotImplementedError()

    def find_in_memory(
        self,
        pattern: bytearray,
        start: int,
        size: int,
        align: int,
        max_matches: int = -1,
        step: int = -1,
    ) -> Generator[int, None, None]:
        """
        Searches for a bit pattern in the memory space of the process. The bit
        pattern can be searched for in a given memory range, and with a given
        alignment. The maximum number of matches that will be generated is
        given by `max_matches`. A value of `max_matches` of `-1` will generate
        all matches.
        """
        raise NotImplementedError()

    def is_remote(self) -> bool:
        """
        Returns whether this process is a remote process connected to using the
        GDB remote debugging protocol.
        """
        raise NotImplementedError()

    def send_remote(self, packet: str) -> bytes:
        """
        Sends the given packet to the GDB remote debugging protocol server.
        Should only be called if `is_remote()` is true.
        """
        raise NotImplementedError()

    def send_monitor(self, cmd: str) -> str:
        """
        Sends the given monitor command to the GDB remote debugging protocol
        server. Should only be called if `is_remote()` is true.
        """
        raise NotImplementedError()

    def download_remote_file(self, remote_path: str, local_path: str) -> None:
        """
        Downloads the given file from the remote host and saves it to the local
        given path. Should only be called if `is_remote()` is true.
        """
        raise NotImplementedError()

    def create_value(self, value: int, type: Type | None = None) -> Value:
        """
        Create a new value in the context of this process, with the given value
        and, optionally, type. If no type is provided, one will be chosen
        automatically.
        """
        raise NotImplementedError()

    # We'll likely have to expand this into a Symbol class and change this to a
    # `symbol_at_address` function later on.
    def symbol_name_at_address(self, address: int) -> str | None:
        """
        Returns the name of the symbol at the given address in the program, if
        one exists.
        """
        raise NotImplementedError()

    def lookup_symbol(
        self,
        name: str,
        *,
        prefer_static: bool = False,
        type: SymbolLookupType = SymbolLookupType.ANY,
        objfile_endswith: str | None = None,
    ) -> Value | None:
        """
        Looks up and returns the address of a symbol by its name.

        Parameters:
        - name (str): The name of the symbol to look up.
        - prefer_static (bool, optional): If True, prioritize symbols in the static block,
          if supported by the debugger. Defaults to False.
        - type (SymbolLookupType, optional): The type of symbol to search for. Defaults
          to SymbolLookupType.ANY.
        - objfile_endswith (str | None, optional): If specified, limits the search to the
          first object file whose name ends with the provided string.

        Returns:
        - pwndbg.dbg_mod.Value | None: The value of the symbol if found, or None if not found.

        Raises:
        - pwndbg.dbg_mod.Error: If no object file matching the `objfile_endswith` pattern is found.
        """

    def get_function_boundaries(self, address: int) -> tuple[int, int] | None:
        """
        Return the function start and end address for a function that
        contains address `addr`.

        Returns:
        - tuple[int, int] | None: [start, end) of function block if found (end address is exclusive)
        """
        raise NotImplementedError()

    # There is an interesting counterpart to this method that exists at the
    # module level. Depending on how we want to implement support for multiple
    # modules, it might be interesting to repeat it there.
    def types_with_name(self, name: str) -> Sequence[Type]:
        """
        Returns a list of all types in this process that match the given name.
        """
        raise NotImplementedError()

    def arch(self) -> ArchDefinition:
        """
        The default architecture of this process.
        """
        raise NotImplementedError()

    def break_at(
        self,
        location: BreakpointLocation | WatchpointLocation,
        stop_handler: Callable[[StopPoint], bool] | None = None,
        internal: bool = False,
    ) -> StopPoint:
        """
        Install a breakpoint or watchpoint at the given location.

        The type of the location determines whether the newly created object
        is a watchpoint or a breakpoint. `BreakpointLocation` locations yield
        breakpoints, while `WatchpointLocation` locations yield watchpoints.

        Aditionally, one may specify a stop handler function, to be run when
        the breakpoint or whatchpoint is hit, and that determines whether
        execution should stop. With a return value of `True` being interpreted
        as a signal to stop, and a return value of `False` being interpreted as
        a signal to continue execution. The extent of the actions that may be
        taken during the stop handler is determined by the debugger.

        Marking a breakpoint or watchpoint as `internal` hints to the
        implementation that the created breakpoint or watchpoint should not be
        directly nameable by the user, and that it should not print any messages
        upon being triggered. Implementations should try to honor this hint,
        but they are not required to in case honoring it is either not possible
        or comes at a significant impact to performance.

        This function returns a handle to the newly created breakpoint or
        watchpoint.
        """
        raise NotImplementedError()

    def trace_ret(self, stop_handler: Callable[[], bool] | None = None, internal: bool = False):
        """
        Traces/break_at the current frame's return address.
        `stop_handler` and `internal` have the same semantic meaning as they are in `break_at`
        """
        raise NotImplementedError()

    # This is a fairly lazy solution. We would ideally support a more robust way
    # to query for ABIs, but Pwndbg currely only uses `show osabi` in GDB to
    # check for whether the target is running under Linux, so we only implement
    # that check.
    def is_linux(self) -> bool:
        """
        Returns whether the current ABI is GNU/Linux.
        """
        raise NotImplementedError()

    def disasm(self, address: int) -> DisassembledInstruction | None:
        """
        Returns the disassembled instruction at the given address in the address
        space of the running process, or `None` if there's no valid instruction
        at that address.
        """
        raise NotImplementedError()

    # We probably want to expose a better module interface in the future, but,
    # for now, this is good enough.
    def module_section_locations(self) -> list[tuple[int, int, str, str]]:
        """
        Return a list of (address, size, section_name, module_name) tuples for
        the loaded sections in every module of this process.

        The module name will have its full path resolved without following symlinks,
        so it is not guaranteed to be the same string as in `/proc/<pid>/maps`
        or vmmap.
        """
        raise NotImplementedError()

    def main_module_name(self) -> str | None:
        """
        Returns the name of the main module.

        On remote targets, this may be prefixed with "target:" string.
        """
        raise NotImplementedError()

    def main_module_entry(self) -> int | None:
        """
        Returns the entry point of the main module.
        """
        raise NotImplementedError()

    def is_dynamically_linked(self) -> bool:
        """
        Returns whether this process makes use of dynamically linked libraries.

        # `"dynamically linked"`
        What exactly it means to be "dynamically linked" here is a little
        ill-defined. Ideally, this function should return true if the process
        uses the default dynamic linker for the system, as that would better
        reflect whether the process uses dynamic linking.

        Currently, though, Pwndbg expects it to behave the same as a check for
        the string "No shared libraries loaded at this time." in the output of
        the `info dll` GDB command, which checks for the presence of other
        modules in the address space of the process, rather than whether or not
        the dynamic linker is used.

        We should probably sort this out in the future.
        """
        raise NotImplementedError()

    def dispatch_execution_controller(
        self, procedure: Callable[[ExecutionController], Coroutine[Any, Any, None]]
    ):
        """
        Queues up the given execution controller-based coroutine for execution,
        sometime between the calling of this function and the
        """
        raise NotImplementedError()

    def add_symbol_file(self, path: str, base: int | None = None) -> None:
        """
        Adds a symbol file at base.
        """
        raise NotImplementedError()

    def remove_symbol_file(self, path: str) -> bool:
        """
        Removes a symbol file.

        Returns:
            True if we succeeded, False if not. If the file was never
            added or doesn't exist, that counts as failure.
        """
        raise NotImplementedError()

    def runcmd(self, cmd: str) -> str:
        """
        Runs a debugger command and returns the output as a string.
        """
        raise NotImplementedError()


class TypeCode(Enum):
    """
    Broad categories of types.
    """

    INVALID = -1
    POINTER = 1
    ARRAY = 2
    STRUCT = 3
    TYPEDEF = 4
    UNION = 5
    INT = 6
    ENUM = 7
    FUNC = 8
    BOOL = 9


class TypeField:
    """
    The fields in a structured type.

    Currently this is just a mirror of `gdb.Field`.
    """

    def __init__(
        self,
        bitpos: int,
        name: str | None,
        type: Type,
        parent_type,
        enumval: int | None = None,
        artificial: bool = False,
        is_base_class: bool = False,
        bitsize: int = 0,
    ) -> None:
        self.bitpos = bitpos
        self.name = name
        self.type = type
        self.parent_type = parent_type
        self.enumval = enumval
        self.artificial = artificial
        self.is_base_class = is_base_class
        self.bitsize = bitsize


class Type:
    """
    Class representing a type in the context of an inferior process.
    """

    @property
    def name_identifier(self) -> str | None:
        """
        Returns the identifier of this type, eg:
        - someStructName
        - someEnumName
        - someTypedefName

        Returns None if the type is anonymous or does not have a name, such as:
        - Anonymous structs
        - Anonymous Typedefs
        - Basic types like char[], void, etc.
        """
        raise NotImplementedError()

    @property
    def name_to_human_readable(self) -> str:
        """
        Returns the human friendly name of this type, eg:
        - char [16]
        - int
        - char *
        - void *
        - fooStructName
        - barEnumName
        - barTypedefName

        This function is not standardized, may return different names in gdb/lldb, eg:
        gdb: `char [16]` or `char [50]` or `struct {...}`
        lldb: `char[16]` or `char[]`    or `(anonymous struct)`

        You should not use this function. Only for human eyes.
        """
        raise NotImplementedError()

    @property
    def array_len(self) -> int:
        """
        Get array length of this type.
        """
        if self.code == pwndbg.dbg_mod.TypeCode.ARRAY:
            return self.sizeof // self.target().sizeof
        return 0

    @property
    def sizeof(self) -> int:
        """
        The size of this type, in bytes.
        """
        raise NotImplementedError()

    @property
    def alignof(self) -> int:
        """
        The alignment of this type, in bytes.
        """
        raise NotImplementedError()

    @property
    def code(self) -> TypeCode:
        """
        What category of type this object belongs to.
        """
        raise NotImplementedError()

    def func_arguments(self) -> list[Type] | None:
        """
        Returns a list of function arguments type.

        Returns:
            List[Type] | None: The function arguments type, or None if debug information is missing.

        Raises:
            TypeError: If called on an unsupported type.
        """
        raise NotImplementedError()

    def fields(self) -> list[TypeField]:
        """
        List of all fields in this type, if it is a structured type.
        """
        raise NotImplementedError()

    def has_field(self, name: str) -> bool:
        """
        Whether this type has a field with the given name.
        """
        # This is a sensible default way to check for a field's existence.
        #
        # Implementations should, however, override this method if there's a
        # debugger-specific check for this that might be faster or more accurate.
        fields = self.fields()
        if fields:
            for field in fields:
                if field.name == name:
                    return True
        return False

    def array(self, count: int) -> Type:
        """
        Return a type that corresponds to an array whose elements have this type.
        """
        raise NotImplementedError()

    def pointer(self) -> Type:
        """
        Return a pointer type that has this type as its pointee.
        """
        raise NotImplementedError()

    def strip_typedefs(self) -> Type:
        """
        Return a type that corresponds to the base type after a typedef chain,
        if this is a typedef. Returns the type itself otherwise.
        """
        raise NotImplementedError()

    def target(self) -> Type:
        """
        Return the target of this reference type, if this is a reference type.
        """
        raise NotImplementedError()

    def keys(self) -> list[str]:
        """
        Returns a list containing all the field names of this type.
        """
        # Like with `has_fields`, we provide a sensible default implementation
        # based on `fields()`. Implementations are encouraged to override this
        # if there is a better debugger-specific way to do this.
        return [field.name for field in self.fields()]

    def enum_member(self, field_name: str) -> int | None:
        """
        Retrieve the integer value of an enum member.

        It returns:
        - integer value, when found field
        - returns None, If the field does not exist
        """
        if self.code != TypeCode.ENUM:
            raise TypeError("only enum supported")

        return next((f.enumval for f in self.fields() if f.name == field_name), None)

    def _offsetof(
        self, field_name: str, *, base_offset_bits: int = 0, nested_cyclic_types: list[Type] = None
    ) -> int | None:
        NESTED_TYPES = (TypeCode.STRUCT, TypeCode.UNION)
        struct_type = self
        if nested_cyclic_types is None:
            nested_cyclic_types = []

        if struct_type.code == TypeCode.TYPEDEF:
            struct_type = struct_type.strip_typedefs()

        if struct_type.code not in NESTED_TYPES:
            return None
        if struct_type in nested_cyclic_types:
            return None

        # note: lldb.SBType and gdb.Type dont support Sets
        nested_cyclic_types.append(struct_type)

        for field in struct_type.fields():
            field_offset_bits = base_offset_bits + field.bitpos

            if field.name == field_name:
                if field_offset_bits % 8 != 0:
                    # Possible bit-fields, misaligned struct, or unexpected alignment
                    # This case is not supported because it introduces complexities
                    # in handling non-byte-aligned or bit-level field offsets
                    return None
                return field_offset_bits // 8

            nested_offset = field.type._offsetof(
                field_name,
                base_offset_bits=field_offset_bits,
                nested_cyclic_types=nested_cyclic_types,
            )
            if nested_offset is not None:
                return nested_offset

        return None

    def offsetof(self, field_name: str) -> int | None:
        """
        Calculate the byte offset of a field within a struct or union.

        This method recursively traverses nested structures and unions, and it computes the
        byte-aligned offset for the specified field.

        It returns:
        - offset in bytes if found
        - None if the field doesn't exist or if an unsupported alignment/bit-field is encountered
        """
        if self.code == TypeCode.POINTER:
            return self.target()._offsetof(field_name)
        return self._offsetof(field_name)

    def __eq__(self, rhs: object) -> bool:
        """
        Returns True if types are the same
        """
        raise NotImplementedError()


class Value:
    """
    Class representing a value in the context of an inferior process.
    """

    @property
    def address(self) -> Value | None:
        """
        The address of this value, in memory, if addressable, otherwise `None`.
        """
        raise NotImplementedError()

    # is_optimized_out is kind of a janky piece of API, honestly. It makes it
    # so that one's ability to call all other methods in this class is often
    # conditional on it being false, and it effectively splits the type into
    # two.
    #
    # There's only _one_ part of Pwndbg that uses it, and I really feel like we
    # should handle variables that have been optimized out some other way.
    #
    # TODO: Remove uses of is_optimized_out from plist and get rid of this.
    @property
    def is_optimized_out(self) -> bool:
        """
        Whether this value is present in debugging information, but has been
        optimized out of the actual program.
        """
        raise NotImplementedError()

    @property
    def type(self) -> Type:
        """
        The type associated with this value.
        """
        raise NotImplementedError()

    def dereference(self) -> Value:
        """
        If this is a pointer value, dereferences the pointer and returns a new
        instance of Value, containing the value pointed to by this pointer.
        """
        raise NotImplementedError()

    # The intent of this function has a great deal of overlap with that of
    # `pwndbg.aglib.memory.string()`. It probably makes sense to take this
    # functionality out of the debugger API.
    #
    # TODO: Move to single, common string function.
    def string(self) -> str:
        """
        If this value is a string, then this method converts it to a Python string.
        """
        raise NotImplementedError()

    def value_to_human_readable(self) -> str:
        """
        Converts a Value to a human-readable string representation.

        The format is similar to what is produced by the `str()` function for gdb.Value,
        displaying nested fields and pointers in a user-friendly way.

        **Usage Notes:**
        - This function is intended solely for displaying results to the user.
        - The output format may differ between debugger implementations (e.g., GDB vs LLDB),
          as each debugger may format values differently. For instance:
            - GDB might produce: '{\n  value = 0,\n  inner = {\n    next = 0x555555558098 <inner_a_node_b+8>\n  }\n}'
            - LLDB might produce: '(inner_a_node) *$PWNDBG_CREATED_VALUE_0 = {\n  value = 0\n  inner = {\n    next = 0x0000555555558098\n  }\n}'
        - As such, this function should not be relied upon for parsing or programmatic use.
        """
        raise NotImplementedError()

    # This is a GDB implementation detail.
    def fetch_lazy(self) -> None:
        """
        Fetches the value if it is lazy, does nothing otherwise.
        """
        raise NotImplementedError()

    def __int__(self) -> int:
        """
        Converts this value to an integer, if possible.
        """
        raise NotImplementedError()

    # Because casting is still sloppy (i.e. it accepts `gdb.Type` objects) in
    # some places, we have to allow `Any` here for lints to pass.
    #
    # TODO: Remove Any type from this function.
    def cast(self, type: Type | Any) -> Value:
        """
        Returns a new value with the same value as this object, but of the
        given type.
        """
        raise NotImplementedError()

    def __add__(self, rhs: int) -> Value:
        """
        Adds an integer to this value, if that makes sense. Throws an exception
        otherwise.
        """
        raise NotImplementedError()

    def __sub__(self, rhs: int) -> Value:
        """
        Subtract an integer from this value, if that makes sense. Throws an
        exception otherwise.
        """
        raise NotImplementedError()

    def __getitem__(self, idx: int | str) -> Value:
        """
        Gets the value with the given name that belongs to this value. For
        structure types, this is the field with the given name. For array types,
        this is the field at the given index. For pointer types, this is the
        value of `*(ptr+idx)`.
        """
        raise NotImplementedError()


class CommandHandle:
    """
    An opaque handle to an installed command.
    """

    def remove(self) -> None:
        """
        Removes this command from the command palette of the debugger.
        """
        raise NotImplementedError()


class EventType(Enum):
    """
    Events that can be listened for and reacted to in a debugger.
    """

    SUSPEND_ALL = -1

    START = 0
    """This event is fired some time between the creation of or attachment to the
    process to be debugged, and the start of its execution."""

    STOP = 1
    """This event is fired after execution of the process has been suspended, but
    before control is returned to the user for interactive debugging."""

    EXIT = 2
    """This event is fired after the process being debugged has been
    detached from or has finished executing."""

    MEMORY_CHANGED = 3
    """This event is fired when the user interactively makes changes to the memory
    of the process being debugged."""

    REGISTER_CHANGED = 4
    """This event is fired when the user interactively makes changes to the registers
    of the process being debugged."""

    CONTINUE = 5
    """This event is fired after the user has requested for process execution to continue
    after it had been previously suspended.

    Be careful about using this event since the debugger may not allow many operations as
    it may consider the program as 'running' when this event is dispatched (e.g. #3683)."""

    NEW_MODULE = 6
    """This event is fired when a new application module has been encountered by the
    debugger. This usually happens when a new application module is loaded into the
    memory space of the process being debugged. In GDB terminology, these are called
    `objfile`s."""


class EventHandlerPriority(Enum):
    """
    Determines the order in which event handlers are called when
    a given event is triggered.
    """

    # If you wish to add another priority, feel free to. Prefer descriptive names
    # like "CACHE_CLEAR" to something generic like "LOW" or "HIGH" which doesn't
    # make it obvious which handlers are registered with this priority.
    # Make sure that the values are defined in increasing order!!
    # (https://docs.python.org/3/library/enum.html#enum.EnumType.__iter__)
    CACHE_CLEAR = 0
    """The first thing we want to do is clear the cache, so we aren't working on stale
    data."""
    UPDATE_ARCH_AND_TYPEINFO = 10
    """We need to initialize the architecture and type information before doing anything
    else substantial."""
    SAVE_SIGNAL = 20
    """Save siginfo information for displaying in the context."""
    STANDARD = 100
    """The default value."""


class Debugger:
    """
    The base class representing a debugger.
    """

    def setup(self, *args: Any) -> None:
        """
        Perform debugger-specific initialization.

        This method should be run immediately after `pwndbg.dbg` is set to an
        instance of this class, and, as such, is allowed to run code that
        depends on it being set.

        Because we can't really know what a given debugger object will need as
        part of its setup process, we allow for as many arguments as desired to
        be passed in, and leave it up to the implementations to decide what they
        need. This shouldn't be a problem, seeing as, unlike other methods in
        this class, this should only be called as part of the debugger-specific
        bringup code.
        """
        raise NotImplementedError()

    def history(self, last: int = 10) -> list[tuple[int, str]]:
        """
        The command history of the interactive session in this debugger.

        This function returns the last `last` items in the command history, as
        an oldest-to-youngest-sorted list of tuples, where the first element in
        each tuple is the index of the command in the history, and the second
        element is a string giving the command itself.
        """
        raise NotImplementedError()

    def lex_args(self, command_line: str) -> list[str]:
        """
        Lexes the given command line into a list of arguments, according to the
        conventions of the debugger being used and of the interactive session.
        """
        raise NotImplementedError()

    def selected_inferior(self) -> Process:
        """
        The inferior process currently being focused on in this interactive session.

        Raises:
            pwndbg.dbg_mod.NoInferior: If the debugger couldn't return an inferior. If
                there is an alive Process (i.e. you are under @pwndbg.commands.OnlyWhenRunning)
                this will not be raised.
        """
        raise NotImplementedError()

    def selected_thread(self) -> Thread | None:
        """
        The thread currently being focused on in this interactive session.
        """
        raise NotImplementedError()

    def selected_frame(self) -> Frame | None:
        """
        The stack frame currently being focused on in this interactive session.
        """
        raise NotImplementedError()

    def commands(self) -> list[str]:
        """
        List the commands available in this session.
        """
        raise NotImplementedError()

    def add_command(
        self,
        name: str,
        handler: Callable[[Debugger, str, bool], None],
        doc: str | None,
        subcommand_names: list[str] | None = None,
    ) -> CommandHandle:
        """
        Adds a command with the given name to the debugger, that invokes the
        given function every time it is called.

        subcommand_names is used for tab-completion.
        """
        raise NotImplementedError()

    def has_event_type(self, ty: EventType) -> bool:
        """
        Whether the given event type is supported by this debugger. Indicates
        that a user either can or cannot register an event handler of this type.
        """
        raise NotImplementedError()

    def event_handler(
        self, event_type: EventType, priority: EventHandlerPriority = EventHandlerPriority.STANDARD
    ) -> Callable[[Callable[..., None]], Callable[..., None]]:
        """
        Sets up the given function to be called when an event of the given type
        gets fired. Returns a callable that corresponds to the wrapped function.
        The wrapped function must return None.

        You may control the order in which the handlers for a specific event will
        be called by setting the `priority` argument.

        This function may be used as a decorator.
        """
        raise NotImplementedError()

    @contextlib.contextmanager
    def ctx_suspend_events(self, ty: EventType) -> Iterator[None]:
        """
        Context manager for temporarily suspending and resuming the delivery of events
        of a given type.
        """

        self.suspend_events(ty)
        try:
            yield
        finally:
            self.resume_events(ty)

    @contextlib.contextmanager
    def ctx_suspend_once(self):
        """
        Avoid printing ctx once
        """
        raise NotImplementedError()

    def suspend_events(self, ty: EventType) -> None:
        """
        Suspend delivery of all events of the given type until it is resumed
        through a call to `resume_events`.

        Events triggered during a suspension will be ignored, and will not be
        delived, even after delivery is resumed.
        """
        raise NotImplementedError()

    def resume_events(self, ty: EventType) -> None:
        """
        Resume the delivery of all events of the given type, if previously
        suspeded through a call to `suspend_events`. Does nothing if the
        delivery has not been previously suspeded.
        """
        raise NotImplementedError()

    def set_sysroot(self, sysroot: str) -> bool:
        """
        Sets the system root for this debugger.
        """
        raise NotImplementedError()

    def x86_disassembly_flavor(self) -> Literal["att", "intel"]:
        """
        The flavor of disassembly to use for x86 targets.
        """
        raise NotImplementedError()

    def supports_breakpoint_creation_during_stop_handler(self) -> bool:
        """
        Whether breakpoint or watchpoint creation through `break_at` is
        supported during breakpoint stop handlers.
        """
        raise NotImplementedError()

    def breakpoint_locations(self) -> list[BreakpointLocation]:
        """
        Returns a list of all breakpoint locations that are currently
        installed and enabled in the focused process.
        """
        raise NotImplementedError()

    # WARNING
    #
    # These are hacky parts of the API that were strictly necessary to bring up
    # pwndbg under LLDB without breaking it under GDB. Expect most of them to be
    # removed or replaced as the porting work continues.
    #

    def name(self) -> DebuggerType:
        """
        The type of the current debugger.
        """
        raise NotImplementedError()

    # We'd like to be able to gate some imports off during porting. This aids in
    # that.
    def is_gdblib_available(self) -> bool:
        """
        Whether gdblib is available under this debugger.
        """
        raise NotImplementedError()

    def string_limit(self) -> int:
        """
        The maximum size of a string.
        """
        raise NotImplementedError()

    def addrsz(self, address: Any) -> str:
        """
        Format the given address value.
        """
        raise NotImplementedError()

    def get_cmd_window_size(self) -> tuple[int, int]:
        """
        The size of the command window, in characters, if available.
        """
        raise NotImplementedError()

    @property
    def pre_ctx_lines(self) -> int:
        """
        Our prediction on how many lines of text will be printed as
        a preamble (right after the prompt, and before the context)
        the next time the context is printed.

        This includes any lines the underlying debugger generates.

        The user never sees these lines when context-clear-screen
        is enabled.
        """
        raise NotImplementedError()

    def set_python_diagnostics(self, enabled: bool) -> None:
        """
        Enables or disables Python diagnostic messages for this debugger.
        """
        raise NotImplementedError()

    def set_convenience_var(self, name: str, value: str, type: str | None) -> None:
        """
        Set a convenience variable which will be accessible with $name in the
        debugger.

        >> Here be dragons. Wear armor. <<

        This is really finicky in LLDB:
        1. It seems convenience variables get an undefined value after process restart.
           (see this comment
            https://github.com/llvm/llvm-project/issues/84806#issuecomment-1995055683)
        2. We cannot "redefine" the variable after it is once created.
        3. Ergo, we cannot change its type.

        Thus for LLDB, we will not honor the `type` parameter and will just set it to void*
        for maximum flexibility. The parameter will be honored for GDB, though.

        Next, be aware that the contents of the `value` variable are passed directly to the
        debugger. So you must pass '"my cool string"' if you actually want the debugger to
        see the quotes.

        Futher, you cannot set convenience variables to some values/types until the process is alive.
        For instance:
            pwndbg> p $wow = ("this is fine")
            $4 = "this is fine"
            pwndbg> p $wow = ((const char*)"this will error")
            evaluation of this expression requires the target program to be active
            pwndbg>

        You should always surround this function with a try/except.
        """
        raise NotImplementedError()



from __future__ import annotations

import argparse
from collections import defaultdict

import pwndbg.color as color
import pwndbg.color.message as message
import pwndbg.commands
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(description="Prints out a list of all Pwndbg commands.")

cat_group = parser.add_mutually_exclusive_group()
cat_group.add_argument(
    "-c", "--category", type=str, default=None, dest="category_", help="Filter commands by category"
)
cat_group.add_argument(
    "--list-categories", dest="list_categories", action="store_true", help="List command categories"
)

parser.add_argument(
    "filter_pattern",
    type=str,
    nargs="?",
    default=None,
    help="Filter to apply to commands names/docs",
)


@pwndbg.commands.Command(parser, command_name="pwndbg", category=CommandCategory.PWNDBG)
def pwndbg_(filter_pattern, category_, list_categories) -> None:
    if list_categories:
        for category in CommandCategory:
            print(color.bold(color.green(f"{category.value}")))
        return

    from tabulate import tabulate

    table_data = defaultdict(list)
    for name, aliases, category, docs in list_and_filter_commands(filter_pattern):
        alias_str = ""
        if aliases:
            aliases = map(color.blue, aliases)
            alias_str = f" [{', '.join(aliases)}]"

        command_names = color.green(name) + alias_str
        table_data[category].append((command_names, docs))

    for category in CommandCategory:
        if category not in table_data or category_ and category_.lower() not in category.lower():
            continue
        data = table_data[category]

        category_header = color.bold(color.green(category + " Commands"))
        alias_header = color.bold(color.blue("Aliases"))
        print(
            tabulate(
                data,
                headers=[f"{category_header} [{alias_header}]", f"{color.bold('Description')}"],
            )
        )
        print()

    print(message.info("Also check out convenience functions with `help function`!"))


def list_and_filter_commands(filter_str: str) -> list[tuple[str, list[str], CommandCategory, str]]:
    """
    Returns:
        A list of (name, aliases, category, description) tuples.
    """
    sorted_commands = sorted(pwndbg.commands.commands, key=lambda c: c.command_name)

    if filter_str:
        filter_str = filter_str.lower()

    results = []

    for c in sorted_commands:
        name = c.command_name
        desc = c.description

        assert desc
        desc = desc.splitlines()[0]

        if not filter_str or filter_str in name.lower() or (desc and filter_str in desc.lower()):
            results.append((name, c.aliases, c.category, desc))

    return results



from __future__ import annotations

import argparse

import pwndbg.aglib
import pwndbg.aglib.memory
import pwndbg.aglib.proc
import pwndbg.commands
from pwndbg.color import message
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(description="Put comments in assembly code.")
parser.add_argument(
    "--addr", metavar="address", default=None, type=str, help="Address to write comments"
)
parser.add_argument("comment", type=str, default=None, help="The text you want to comment")

file_lists: dict[str, dict[str, str]] = {}  # This saves all comments.


@pwndbg.commands.Command(parser, category=CommandCategory.MISC)
@pwndbg.commands.OnlyWhenRunning
def comm(addr=None, comment=None) -> None:
    if addr is None:
        addr = hex(pwndbg.aglib.regs.pc)
    try:
        with open(".gdb_comments", "a+") as f:
            target = int(addr, 0)

            if not pwndbg.aglib.memory.peek(target):
                print(message.error(f"Invalid Address {target:#x}"))

            else:
                f.write(f"file:{pwndbg.aglib.proc.exe()}=")
                f.write(f"{target:#x}:{comment}\n")
                if pwndbg.aglib.proc.exe() not in file_lists:
                    file_lists[pwndbg.aglib.proc.exe()] = {}
                file_lists[pwndbg.aglib.proc.exe()][hex(target)] = comment
    except Exception:
        print(message.error("Permission denied to create file"))


def init() -> None:
    try:
        with open(".gdb_comments") as f:
            text = f.read()
            text = text.split("\n")
            for i in range(len(text) - 1):
                text1, text2 = text[i].split("=")

                # split Filename, comments
                filename = text1.split(":")[1]
                addr_comm = text2.split(":")

                if filename not in file_lists:
                    file_lists[filename] = {}

                file_lists[filename][addr_comm[0]] = addr_comm[1]

    except Exception:
        pass



"""
Determine whether the target is being run under QEMU.
"""

from __future__ import annotations

import pwndbg
import pwndbg.lib.cache
import pwndbg.lib.qemu


@pwndbg.lib.cache.cache_until("stop")
def qemu_gdbserver_version() -> tuple[int, ...] | None:
    """
    Returns QEMU version. Works since QEMU 10.1.0
    """
    inferior = pwndbg.dbg.selected_inferior()
    if not inferior.is_remote():
        return None

    response = inferior.send_remote("qGDBServerVersion")
    return pwndbg.lib.qemu.parse_qgdbserverversion(response)


@pwndbg.lib.cache.cache_until("stop")
def is_qemu() -> bool:
    inferior = pwndbg.dbg.selected_inferior()
    if not inferior.is_remote():
        return False

    # Examples:
    #
    # pwndbg> maintenance packet Qqemu.sstepbits
    # sending: "Qqemu.sstepbits"
    # received: "ENABLE=1,NOIRQ=2,NOTIMER=4"
    #
    # pwndbg-lldb> process plugin packet send Qqemu.sstepbits
    #   packet: Qqemu.sstepbits
    # response: ENABLE=1,NOIRQ=2,NOTIMER=4
    #
    response = inferior.send_remote("Qqemu.sstepbits")

    return b"ENABLE=" in response


@pwndbg.lib.cache.cache_until("stop")
def is_usermode() -> bool:
    inferior = pwndbg.dbg.selected_inferior()
    if not inferior.is_remote():
        return False

    # There is also 'qAttached' - maybe we can use it too?
    # for qemu user though it returned "0"?
    # Try with:
    #    qemu-x86_64 -g 1234 `which ps`
    #    gdb -nx `which ps` -ex 'target remote :1234'
    response = inferior.send_remote("qOffsets")

    return b"Text=" in response


@pwndbg.lib.cache.cache_until("stop")
def is_qemu_usermode() -> bool:
    """Returns ``True`` if the target remote is being run under
    QEMU usermode emulation."""

    return is_qemu() and is_usermode()


@pwndbg.lib.cache.cache_until("stop")
def is_qemu_kernel() -> bool:
    return is_qemu() and not is_usermode()


def is_old_qemu_user() -> bool:
    # qemu-user <8.1
    return is_qemu_usermode() and not exec_file_supported()


@pwndbg.lib.cache.cache_until("stop")
def exec_file_supported() -> bool:
    """Returns ``True`` if the remote target understands the 'qXfer:exec-file:read' packet.
    A check for this feature is done in vmmap code, to warn against running legacy Qemu versions.
    """
    response = pwndbg.dbg.selected_inferior().send_remote("qSupported")

    return b"qXfer:exec-file:read" in response



from __future__ import annotations

from collections.abc import Iterator

import pwndbg
import pwndbg.aglib.memory
import pwndbg.aglib.typeinfo
import pwndbg.dbg_mod
from pwndbg.aglib.kernel.macros import container_of
from pwndbg.dbg_mod import EventType

rb_root_type: pwndbg.dbg_mod.Type = None
rb_node_type: pwndbg.dbg_mod.Type = None


@pwndbg.dbg.event_handler(EventType.START)
def init():
    global rb_root_type, rb_node_type
    rb_root_type = pwndbg.aglib.typeinfo.load("struct rb_root")
    rb_node_type = pwndbg.aglib.typeinfo.load("struct rb_node")


def for_each_rb_entry(
    root: pwndbg.dbg_mod.Value, typename: str, fieldname: str
) -> Iterator[pwndbg.dbg_mod.Value]:
    node = rb_first(root)
    node_addr = int(node or 0)
    while node_addr != 0:
        yield container_of(node_addr, typename, fieldname)
        node = rb_next(node)
        node_addr = int(node or 0)


def rb_first(root: pwndbg.dbg_mod.Value) -> pwndbg.dbg_mod.Value | None:
    if root.type == rb_root_type:
        node = root.address.cast(rb_root_type.pointer())
    elif root.type != rb_root_type.pointer():
        raise pwndbg.dbg_mod.Error(f"Must be struct rb_root not {root.type}")

    node = root["rb_node"]
    if int(node) == 0:
        return None

    while int(node["rb_left"]):
        node = node["rb_left"]

    return node


def rb_last(root: pwndbg.dbg_mod.Value) -> pwndbg.dbg_mod.Value | None:
    if root.type == rb_root_type:
        node = root.address.cast(rb_root_type.pointer())
    elif root.type != rb_root_type.pointer():
        raise pwndbg.dbg_mod.Error(f"Must be struct rb_root not {root.type}")

    node = root["rb_node"]
    if int(node) == 0:
        return None

    while int(node["rb_right"]):
        node = node["rb_right"]

    return node


def rb_parent(node: pwndbg.dbg_mod.Value) -> pwndbg.dbg_mod.Value:
    val = int(node["__rb_parent_color"]) & ~3
    return pwndbg.aglib.memory.get_typed_pointer(rb_node_type, val)


def rb_empty_node(node: pwndbg.dbg_mod.Value) -> bool:
    return int(node["__rb_parent_color"]) == int(node.address or 0)


def rb_next(node: pwndbg.dbg_mod.Value) -> pwndbg.dbg_mod.Value | None:
    if node.type == rb_node_type:
        node = node.address.cast(rb_node_type.pointer())
    elif node.type != rb_node_type.pointer():
        raise pwndbg.dbg_mod.Error(f"Must be struct rb_node not {node.type}")

    if rb_empty_node(node):
        return None

    if int(node["rb_right"]):
        node = node["rb_right"]
        while int(node["rb_left"]):
            node = node["rb_left"]
        return node

    parent = rb_parent(node)
    while int(parent) and int(node) == int(parent["rb_right"]):
        node = parent
        parent = rb_parent(node)

    return parent


def rb_prev(node: pwndbg.dbg_mod.Value) -> pwndbg.dbg_mod.Value | None:
    if node.type == rb_node_type:
        node = node.address.cast(rb_node_type.pointer())
    elif node.type != rb_node_type.pointer():
        raise pwndbg.dbg_mod.Error(f"Must be struct rb_node not {node.type}")

    if rb_empty_node(node):
        return None

    if int(node["rb_left"]):
        node = node["rb_left"]
        while int(node["rb_right"]):
            node = node["rb_right"]
        return node.dereference()

    parent = rb_parent(node)
    while int(parent) and int(node) == int(parent["rb_left"].dereference()):
        node = parent
        parent = rb_parent(node)

    return parent



"""
Looking up addresses for function names / symbols, and
vice-versa.
"""

from __future__ import annotations

import pwndbg.aglib.memory
import pwndbg.dbg_mod
import pwndbg.dintegration
import pwndbg.lib.cache
from pwndbg.dbg_mod import SymbolLookupType


def lookup_symbol_addr(
    name: str,
    *,
    prefer_static: bool = False,
    type: SymbolLookupType = SymbolLookupType.ANY,
    objfile_endswith: str | None = None,
) -> int | None:
    addr = lookup_symbol(
        name, type=type, prefer_static=prefer_static, objfile_endswith=objfile_endswith
    )
    if not addr:
        return None
    return int(addr)


def lookup_symbol_value(
    name: str,
    *,
    prefer_static: bool = False,
    type: SymbolLookupType = SymbolLookupType.ANY,
    objfile_endswith: str | None = None,
) -> int | None:
    addr = lookup_symbol(
        name, type=type, prefer_static=prefer_static, objfile_endswith=objfile_endswith
    )
    if not addr or not pwndbg.aglib.memory.peek(int(addr)):
        return None

    value = addr.dereference()
    if not value:
        return None
    return int(value)


# TODO: cache here? can we cache obj 'Value'?
def lookup_symbol(
    name: str,
    *,
    prefer_static: bool = False,
    type: SymbolLookupType = SymbolLookupType.ANY,
    objfile_endswith: str | None = None,
) -> pwndbg.dbg_mod.Value | None:
    """
    Returns the address of the given `symbol`, cast-ed to the appropriate symbol type.

    This function searches for (SymbolLookupType.ANY):
    - Function names
    - Variable names
    - (gdb only) Typedef names (if you need please use aglib.typeinfo.load)
    - (gdb only) Enum values (if you need please use aglib.typeinfo.enum_member)

    The lookup order is as follows (default):
    1. Global scope within the current module
    2. Global static scope within the current module
    3. Global scope in other modules
    4. Global static scope in other modules
    """
    return pwndbg.dbg.selected_inferior().lookup_symbol(
        name, type=type, prefer_static=prefer_static, objfile_endswith=objfile_endswith
    )


# TODO: cache here? can we cache obj 'Value'?
def lookup_frame_symbol(
    name: str, *, type: SymbolLookupType = SymbolLookupType.ANY
) -> pwndbg.dbg_mod.Value | None:
    """
    Returns the address of the given `symbol`, cast-ed to the appropriate symbol type.

    This function searches for (SymbolLookupType.ANY):
    - Function names
    - Variable names
    - (gdb only, please don't use) Typedef names
    - (gdb only, please don't use) Enum values

    The lookup order is as follows:
    1. Local scope
    2. Global scope within the current module
    3. Global static scope within the current module
    4. Global scope in other modules
    5. Global static scope in other modules
    """
    return pwndbg.dbg.selected_frame().lookup_symbol(name, type=type)


@pwndbg.lib.cache.cache_until("objfile")
def resolve_addr(addr: int) -> str | None:
    """
    Resolves an address to its corresponding symbol name, if available.

    Note:
    - This function does not resolve TLS (Thread-Local Storage) addresses or addresses with local scope.

    Resolution is performed in the following order:
    - Global scope symbols.
    """
    assert addr >= 0, "address must be positive"

    symbol_name = pwndbg.dbg.selected_inferior().symbol_name_at_address(addr)
    if symbol_name:
        return symbol_name

    return pwndbg.dintegration.manager.symbol_at_address(addr)


@pwndbg.lib.cache.cache_until("objfile")
def resolve_function_boundaries(addr: int) -> tuple[int, int] | None:
    """
    Return the function start and end address for a function that
    contains address `addr`.

    Returns:
    - tuple[int, int] | None: [start, end) of function block if found (end address is exclusive)
    """
    assert addr >= 0, "address must be positive"

    return pwndbg.dbg.selected_inferior().get_function_boundaries(addr)



from __future__ import annotations

from enum import Enum

import pwndbg.wrappers

cmd_name = "readelf"


class RelocationType(Enum):
    # For x86_64, some details about these flag can be found in 4.4.1 Relocation Types in https://www.intel.com/content/dam/develop/external/us/en/documents/mpx-linux64-abi.pdf
    # The definitions of these flags can be found in this file: https://elixir.bootlin.com/glibc/glibc-2.37/source/elf/elf.h
    JUMP_SLOT = 1  # e.g.: R_X86_64_JUMP_SLOT
    GLOB_DAT = 2  # e.g.: R_X86_64_GLOB_DAT
    IRELATIVE = 3  # e.g.: R_X86_64_IRELATIVE


@pwndbg.wrappers.OnlyWithCommand(cmd_name)
def get_got_entry(local_path: str) -> dict[RelocationType, list[str]]:
    # --wide is for showing the full information, e.g.: R_X86_64_JUMP_SLOT instead of R_X86_64_JUMP_SLO
    readelf_out = pwndbg.wrappers.call_cmd(cmd_name, "--relocs", "--wide", local_path)

    entries: dict[RelocationType, list[str]] = {category: [] for category in RelocationType}
    for line in readelf_out.splitlines():
        if not line or not line[0].isdigit() or " " not in line:
            continue
        category = line.split()[2]
        # TODO/FIXME: There's a bug here, somehow the IRELATIVE relocation might point to somewhere in .data.rel.ro, which is not in .got or .got.plt
        for c in RelocationType:
            if c.name in category:
                entries[c].append(line)
    return entries



from __future__ import annotations

import argparse
import subprocess
import tempfile

import gdb

import pwndbg.aglib.proc
import pwndbg.commands
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(
    description="ROP gadget search with ropper.",
)
parser.add_argument("argument", nargs="*", type=str, help="Arguments to pass to ropper")


@pwndbg.commands.Command(
    parser,
    category=CommandCategory.INTEGRATIONS,
    examples="""
pwndbg> ropper -- --console
[INFO] Load gadgets from cache
[LOAD] loading... 100%
[LOAD] removing double gadgets... 100%
(bash/ELF/x86_64)>
----
pwndbg> ropper -- --search 'pop rdi; ret;'
[INFO] Load gadgets from cache
[LOAD] loading... 100%
[LOAD] removing double gadgets... 100%
[INFO] Searching for gadgets: pop rdi; ret;

[INFO] File: /usr/bin/bash
0x000000000001ee23: pop rdi; ret;
    """,
)
@pwndbg.commands.OnlyWithFile
def ropper(argument) -> None:
    with tempfile.NamedTemporaryFile() as corefile:
        # If the process is running, dump a corefile so we get actual addresses.
        if pwndbg.aglib.proc.alive():
            filename = corefile.name
            gdb.execute(f"gcore {filename}")
        else:
            filename = pwndbg.aglib.proc.exe()

        # Build up the command line to run
        cmd = ["ropper", "--file", filename]
        cmd += argument

        try:
            subprocess.call(cmd)
        except Exception:
            print("Could not run ropper.  Please ensure it's installed and in $PATH.")



from __future__ import annotations

from os import environ
from typing import Any

import gdb

import pwndbg
import pwndbg.aglib.proc
import pwndbg.commands
import pwndbg.commands.context
import pwndbg.dbg_mod
import pwndbg.decorators
import pwndbg.gdblib.events
import pwndbg.gdblib.functions
import pwndbg.profiling
from pwndbg.color import disable_colors
from pwndbg.color import message
from pwndbg.dbg_mod import EventType
from pwndbg.lib.tips import color_tip
from pwndbg.lib.tips import get_tip_of_the_day

# noinspection PyPackageRequirements
show_tip = pwndbg.config.add_param(
    "show-tips", True, "whether to display the tip of the day on startup"
)

cur: tuple[gdb.Inferior, gdb.InferiorThread] | None = None


def initial_hook(*a: Any) -> None:
    if show_tip and not pwndbg.decorators.first_prompt:
        colored_tip = color_tip(get_tip_of_the_day(pwndbg.dbg_mod.DebuggerType.GDB.value))
        print(
            message.prompt("------- tip of the day")
            + message.system(" (disable with {})".format(message.notice("set show-tips off")))
            + message.prompt(" -------")
        )
        print(colored_tip)
    pwndbg.decorators.first_prompt = True

    prompt_hook(*a)

    if environ.get("PWNDBG_PROFILE") == "1":
        pwndbg.profiling.profiler.stop("pwndbg-first-prompt.pstats")

    gdb.prompt_hook = prompt_hook


context_shown = False
last_alive_state = False


def show_hint() -> None:
    hint_lines = (
        f"loaded {len(pwndbg.commands.commands)} pwndbg commands."
        f" Type {message.notice('pwndbg [filter]')} for a list.",
        f"created {len(pwndbg.gdblib.functions.functions)} GDB functions (can be used"
        f" with print/break). Type {message.notice('help function')} to see them.",
    )

    for line in hint_lines:
        print(message.prompt("pwndbg: ") + message.system(line))


def thread_is_stopped() -> bool:
    """
    This detects whether selected thread is stopped.
    It is not stopped in situations when gdb is executing commands
    that are attached to a breakpoint by `command` command.

    For more info see issue #229 ( https://github.com/pwndbg/pwndbg/issues/299 )
    :return: Whether gdb executes commands attached to bp with `command` command.
    """
    t = gdb.selected_thread()
    if not t:
        return False
    return t.is_stopped()


def prompt_hook(*a: Any) -> None:
    global cur, context_shown, last_alive_state

    new = (gdb.selected_inferior(), gdb.selected_thread())

    if cur != new:
        pwndbg.gdblib.events.after_reload(fire_start=cur is None)
        cur = new

    if not context_shown and pwndbg.aglib.proc.alive() and thread_is_stopped():
        pwndbg.commands.context.selected_history_index = None
        pwndbg.commands.context.context()
        context_shown = True

    # set prompt again when alive state changes
    if last_alive_state != pwndbg.aglib.proc.alive():
        last_alive_state = pwndbg.aglib.proc.alive()
        set_prompt()


@pwndbg.dbg.event_handler(EventType.CONTINUE)
def reset_context_shown(*a: Any) -> None:
    global context_shown
    context_shown = False


@pwndbg.config.trigger(message.config_prompt_color, disable_colors)
def set_prompt() -> None:
    prompt = "pwndbg> "

    if not disable_colors:
        if pwndbg.aglib.proc.alive():
            prompt = message.readline_escape(message.alive_prompt, prompt)
        else:
            prompt = message.readline_escape(message.prompt, prompt)

    gdb.execute(f"set prompt {prompt}")


gdb.prompt_hook = initial_hook



from __future__ import annotations

import argparse

import gdb

import pwndbg.commands
from pwndbg.color import message
from pwndbg.commands import CommandCategory
from pwndbg.gdblib.scheduler import lock_scheduler

parser = argparse.ArgumentParser(
    description="""Kill all or given threads.

Switches to given threads and calls pthread_exit(0) on them.
This is performed with scheduler-locking to prevent other threads from operating at the same time.

Killing all other threads may be useful to use GDB checkpoints, e.g., to test given input & restart the execution to the point of interest (checkpoint).
""",
)

parser.add_argument("thread_ids", type=int, nargs="*", help="Thread IDs to kill.")
parser.add_argument(
    "-a",
    "--all",
    action="store_true",
    help="Kill all threads except the current one.",
)


@pwndbg.commands.Command(parser, category=CommandCategory.PROCESS)
@pwndbg.commands.OnlyWhenRunning
def killthreads(thread_ids: list[int] | None = None, all: bool = False) -> None:
    if len(thread_ids) == 0 and not all:
        print(message.error("No thread IDs or --all flag specified"))
        return

    if all and len(thread_ids) > 0:
        print(message.error("Cannot specify thread IDs and --all"))
        return

    with lock_scheduler():
        current_thread_id = gdb.selected_thread().num
        available_thread_ids = [
            thread.num
            for thread in gdb.selected_inferior().threads()
            if thread.num != current_thread_id
        ]
        if all:
            thread_ids = available_thread_ids
        else:
            for thread_id in thread_ids:
                if thread_id not in available_thread_ids:
                    print(
                        message.error(f"Thread ID {thread_id} does not exist, see `info threads`")
                    )
                    return
        for thread_id in thread_ids:
            gdb.execute(f"thread {thread_id}", to_string=True)
            try:
                gdb.execute("call (void) pthread_exit(0)", to_string=True)
            except gdb.error:
                # gdb will throw an error, because the thread dies during the call, which is expected
                pass

        # Switch back to the thread we were on before killing threads
        gdb.execute(f"thread {current_thread_id}", to_string=True)
        print(
            message.success(
                "Killed threads with IDs: "
                + ", ".join([str(thread_id) for thread_id in thread_ids])
            )
        )



from __future__ import annotations

from pwndbg.color import theme

offset_color = theme.add_color_param(
    "telescope-offset-color", "normal", "color of the telescope command (offset prefix)"
)
register_color = theme.add_color_param(
    "telescope-register-color", "bold", "color of the telescope command (register)"
)
offset_separator_color = theme.add_color_param(
    "telescope-offset-separator-color",
    "normal",
    "color of the telescope command (offset separator)",
)
offset_delimiter_color = theme.add_color_param(
    "telescope-offset-delimiter-color",
    "normal",
    "color of the telescope command (offset delimiter)",
)
repeating_marker_color = theme.add_color_param(
    "telescope-repeating-marker-color",
    "normal",
    "color of the telescope command (repeating values marker)",
)


def offset(x: object) -> str:
    return offset_color.color_function(x)


def register(x: object) -> str:
    return register_color.color_function(x)


def separator(x: object) -> str:
    return offset_separator_color.color_function(x)


def delimiter(x: object) -> str:
    return offset_delimiter_color.color_function(x)


def repeating_marker(x: object) -> str:
    return repeating_marker_color.color_function(x)



"""
Pwndbg command implementations.

As well as various command-handling logic.
"""

from __future__ import annotations

import argparse
import functools
import logging
from collections.abc import Callable
from enum import Enum
from typing import Any
from typing import TypeVar

from typing_extensions import ParamSpec
from typing_extensions import override

import pwndbg.aglib
import pwndbg.aglib.heap
import pwndbg.aglib.kernel
import pwndbg.aglib.proc
import pwndbg.aglib.qemu
import pwndbg.aglib.symbol
import pwndbg.color.message as message
import pwndbg.dbg_mod
import pwndbg.dintegration
import pwndbg.exception
import pwndbg.libc
from pwndbg.aglib.heap.ptmalloc import DebugSymsHeap
from pwndbg.aglib.heap.ptmalloc import GlibcMemoryAllocator
from pwndbg.aglib.heap.ptmalloc import HeuristicHeap
from pwndbg.lib import SymbolNotRecoveredError
from pwndbg.lib import TypeNotRecoveredError

log = logging.getLogger(__name__)

T = TypeVar("T")
P = ParamSpec("P")

commands: list[CommandObj] = []
command_names: set[str] = set()


class CommandCategory(str, Enum):
    START = "Start"
    NEXT = "Step/Next/Continue"
    CONTEXT = "Context"
    PTMALLOC2 = "GLibc ptmalloc2 Heap"
    JEMALLOC = "jemalloc Heap"
    BREAKPOINT = "Breakpoint"
    MEMORY = "Memory"
    MUSL = "musl"
    STACK = "Stack"
    REGISTER = "Register"
    PROCESS = "Process"
    LINUX = "Linux/libc/ELF"
    DARWIN = "Darwin/libsystem/Mach-O"
    DISASS = "Disassemble"
    MISC = "Misc"
    KERNEL = "Kernel"
    INTEGRATIONS = "Integrations"
    WINDBG = "WinDbg"
    PWNDBG = "Pwndbg"
    SHELL = "Shell"
    DEV = "Developer"


GDB_BUILTIN_COMMANDS = pwndbg.dbg.commands()

# Set in `reload` command so that we can skip double checking for registration
# of an already existing command when re-registering GDB CLI commands
# (there is no way to unregister a command in GDB 12.x)
pwndbg_is_reloading = False
if pwndbg.dbg.is_gdblib_available():
    import gdb

    pwndbg_is_reloading = getattr(gdb, "pwndbg_is_reloading", False)


class InvalidDebuggerError(Exception):
    """
    Raised when a command is called in a debugger for which
    it is disallowed.
    """

    pass


class CommandFormatter(argparse.RawDescriptionHelpFormatter):
    """
    The formatter_class that is passed to argparse for all
    commands.

    Subclassing this isn't officially supported, but there
    isn't a good alternative.
    """

    @override
    def format_help(self) -> str:
        """
        Formats the help string to reorder it, so that its first description line is first
        and the usage string is second. This means we change the help from:
            usage: command [-flags]

            First line description

            positional arguments: (... etc)

        To:
            First line description

            usage: command [-flags]

            positional arguments: (... etc)

        We do this for GDB as it takes the first line of command help for its 'apropos <cmd>' command.
        See #3502 for more information.
        """

        # Do this only if there are at least two items
        if len(self._root_section.items) >= 2:
            self._root_section.items[0], self._root_section.items[1] = (
                self._root_section.items[1],
                self._root_section.items[0],
            )

        return super().format_help()

    @override
    def _get_help_string(self, action: argparse.Action) -> str:
        # Yoinked from argparse.ArgumentDefaultsHelpFormatter with
        # the added ` and action.default not in (None, False)` check.
        help_ = action.help
        if help_ is None:
            help_ = ""

        if "%(default)" not in help_:
            is_false_bool = (
                action.type is bool or isinstance(action.default, bool)
            ) and not action.default
            is_none = action.default is None
            if action.default is not argparse.SUPPRESS and not (is_false_bool or is_none):
                defaulting_nargs = [argparse.OPTIONAL, argparse.ZERO_OR_MORE]
                if action.option_strings or action.nargs in defaulting_nargs:
                    if action.type is str:
                        help_ += " (default: '%(default)s')"
                    else:
                        help_ += " (default: %(default)s)"

        return help_


class CommandObj:
    """
    Represents a command that can be invoked from the
    debugger.
    """

    builtin_override_whitelist: set[str] = {
        "up",
        "down",
        "search",
        "pwd",
        "start",
        "starti",
        "ignore",
    }
    history: dict[int, str] = {}

    def __init__(
        self,
        function: Callable[..., str | None],
        parser: argparse.ArgumentParser,
        command_name: str | None,
        category: CommandCategory,
        aliases: list[str],
        examples: str,
        notes: str,
        /,  # All parameters must be passed in positionally
    ) -> None:
        assert function
        self.function: Callable[..., str | None] = function

        if command_name is None:
            # Take the command name from the name of the function
            # which defines it, but replace '_' with '-'.
            self.command_name: str = function.__name__.replace("_", "-")
        else:
            self.command_name = command_name

        assert "_" not in self.command_name and "Use '-' instead of '_' in command names."
        assert self.command_name not in command_names and "Command already exists."
        assert (
            not (
                self.command_name in GDB_BUILTIN_COMMANDS
                and self.command_name not in CommandObj.builtin_override_whitelist
                and not pwndbg_is_reloading
            )
            and "Cannot override non-whitelisted built-in command."
        )

        assert category
        self.category: CommandCategory = category

        self.aliases: list[str] = aliases
        self.examples: str = examples.strip()
        self.notes: str = notes.strip()

        assert parser
        self.parser: argparse.ArgumentParser = parser
        # Sets self.help_str, self.description and self.subcommand_names (among other stuff).
        self.help_str: str
        self.description: str
        self.subcommand_names: list[str] | None
        self.initialize_parser()

        # Let the debugger and pwndbg global state know about it.
        self.register_command()

        # For commands like hexdump where you get new output from
        # continuous invocations.
        self.repeat: bool = False

    def register_command(self) -> None:
        """
        Register this object command with the underlying debugger
        and update pwndbg global state to know about this command.
        """

        def _handler(
            _debugger: pwndbg.dbg_mod.Debugger, arguments: str, is_interactive: bool
        ) -> None:
            self.invoke(arguments, is_interactive)

        if self.subcommand_names is not None and len(self.subcommand_names) > 0:
            # In order to add `help <main> <sub>` support, the main
            # command needs to be registered as a prefix command in
            # GDB. Since this causes help info duplication, for now
            # we simply show a hint to use `--help`
            potential_newline: str = "" if self.aliases else "\n"
            self.help_str += f"{potential_newline}Hint: Use `{self.command_name} <subcmd> --help` if you want to see subcommand information."

        # Keep a handle to the command and its aliases so we can
        # easily remove them if necessary (not supported with GDB).
        self.handles = [
            # Tell the debugger about the command...
            pwndbg.dbg.add_command(
                self.command_name, _handler, self.help_str, self.subcommand_names
            )
        ]

        # ...and all of its aliases.
        self.handles.extend(
            pwndbg.dbg.add_command(alias, _handler, self.help_str, self.subcommand_names)
            for alias in self.aliases
        )

        command_names.add(self.command_name)
        commands.append(self)

    @staticmethod
    def has_notes_string(text: str) -> bool:
        return any(nt in text.lower() for nt in ("note:", "notes:"))

    @staticmethod
    def has_examples_string(text: str) -> bool:
        return any(ex in text.lower() for ex in ("example:", "examples:"))

    def setup_epilog(self) -> None:
        # Build the actual epilog from the examples, notes and passed epilog.
        self.epilog = ""
        self.pure_epilog = ""

        if self.examples:
            assert (
                not self.has_examples_string(self.examples)
                and "No need, `Examples:` is added automatically."
            )
            # Not putting '\n' in the notice() so .strip() works properly.
            self.epilog += "\n" + message.notice("Examples:") + "\n"
            self.epilog += self.examples + "\n"

        if self.notes:
            assert (
                not self.has_notes_string(self.notes)
                and "No need, `Notes:` is added automatically."
            )
            self.epilog += "\n" + message.notice("Notes:") + "\n"
            self.epilog += self.notes + "\n"

        if self.parser.epilog:
            self.pure_epilog = self.parser.epilog.strip()
            assert (
                not self.has_examples_string(self.pure_epilog)
                and "Put examples into pwndbg.commands.Command(examples=your_example)."
            )
            assert (
                not self.has_notes_string(self.pure_epilog)
                and "Put notes into pwndbg.commands.Command(notes=your_note)."
            )
            self.epilog += "\n" + self.pure_epilog + "\n"

        if self.aliases:
            alias_txt = "Alias" + ("es" if len(self.aliases) > 1 else "") + ": "
            self.epilog += "\n" + message.notice(alias_txt)
            self.epilog += ", ".join(self.aliases) + "\n"

        # Update the parser so the help is correctly generated.
        self.parser.epilog = self.epilog = self.epilog.strip()

    @staticmethod
    def initialize_parser_recursively(
        parser: argparse.ArgumentParser, top_level_name: str, level: int
    ) -> None:
        if level == 0:
            # Top level command
            assert parser.prog[0] != " "
            assert top_level_name == ""
        else:
            # Workaround until https://github.com/pwndbg/pwndbg/issues/3523
            # is fixed.
            parser.prog = (
                parser.prog.replace("pwndbg-lldb", "")
                .replace("launch_guest.py", "")
                .replace("python3 -m tests.host.lldb.launch_guest", "")
            )
            # A level one subcommand will have parser.prog == " install"
            # while a level two subcommand will have parser.prog == "install ida".
            # Except on lldb, where its " install ida" (after the replace).
            # How does this make sense? So annoying..
            assert top_level_name != ""
            if level == 1:
                assert parser.prog[0] == " ", (
                    "Pwndbg automatically sets the subparser's prog. Don't touch it, just set the name."
                )
            else:
                parser.prog = parser.prog.strip()
                assert parser.prog.count(" ") == level - 1, (
                    "Pwndbg automatically sets the subparser's prog. Don't touch it, just set the name."
                )
                parser.prog = " " + parser.prog

        parser.prog = top_level_name + parser.prog

        # We want to run all integer and otherwise-unspecified arguments
        # through fix() so that GDB parses it.
        for action in parser._actions:
            if action.dest == "help":
                # The HelpAction exists by default and handles `-h` and `--help`.
                # No need to do anything about it.
                continue

            if not isinstance(action, argparse._SubParsersAction) and action.help is None:
                # When we do `cmd -h` we want each argument to have a one-line
                # description.
                # Unfortunately, I don't know how to enforce that each subcommand has a help=
                # passed to its add_parser() :(
                print(message.error(f"Error parsing arguments for command: {parser.prog}"))
                print("You must add a `help=` string to your argument.")
                print(f"Erroneous action:\n\t{repr(action)}\n")
                assert False, "You must add a `help=` string to your argument."

            if action.type is int:
                action.type = fix_int_reraise_arg
            elif type(action) is argparse._StoreAction and action.type is None:
                # Prevents bugs like https://github.com/pwndbg/pwndbg/pull/3477
                print(message.error(f"Error parsing arguments for command: {parser.prog}"))
                print("You must set the argument type for a store action.")
                print(f"Erroneous action:\n\t{repr(action)}\n")
                assert False, "You must set the argument type for a store action."

        assert (
            parser.formatter_class is argparse.HelpFormatter
            and "All pwndbg commands should use the same formatter."
        )

        parser.formatter_class = CommandFormatter

        # Used by `pwndbg [filter]`
        assert (
            parser.description
            and parser.description.strip()
            and "A command must contain a description."
        )
        parser.description = parser.description.strip()

        # Run recursively on subparsers (if any)
        for action in parser._actions:
            if isinstance(action, argparse._SubParsersAction):
                if level == 0:
                    top_level_name = parser.prog
                last_prog = "<doesn't exist>"
                for subparser in action.choices.values():
                    # Argparse creates duplicate objects for aliases, we don't need to
                    # reparse them (and shouldn't, as we will mess up the parser.prog).
                    if subparser.prog != last_prog:
                        CommandObj.initialize_parser_recursively(
                            subparser, top_level_name, level + 1
                        )

                    last_prog = subparser.prog

    def initialize_parser(self) -> None:
        # Set parser.prog so the help is generated properly.
        self.parser.prog = self.command_name

        # Clean up and check subcommands as well
        CommandObj.initialize_parser_recursively(self.parser, "", 0)

        # Add non-alias subcommands to self.subcommand_names which will
        # register them for tab-completion in the debugger.
        self.subcommand_names = None

        for action in self.parser._actions:
            if isinstance(action, argparse._SubParsersAction):
                self.subcommand_names = []
                last_prog = "<doesn't exist>"
                for subcmd_name, subparser in action.choices.items():
                    if subparser.prog != last_prog:
                        self.subcommand_names.append(subcmd_name)
                    last_prog = subparser.prog
                # Not sure what multiple subparser actions would mean..
                break

        assert self.parser.description
        self.description = self.parser.description

        assert (
            not self.has_examples_string(self.description)
            and "Put examples into pwndbg.commands.Command(examples=your_example)."
        )
        assert (
            not self.has_notes_string(self.description)
            and "Put notes into pwndbg.commands.Command(notes=your_note)."
        )

        self.setup_epilog()

        # Generate command help (after stripping the parser's variables
        # and defining a formatter).
        self.help_str = self.parser.format_help()

    def invoke(self, argument: str, from_tty: bool) -> None:
        """Invoke the command with an argument string"""
        try:
            _ = pwndbg.dbg.selected_inferior()
        except pwndbg.dbg_mod.NoInferior:
            log.error("Pwndbg commands require a target binary to be selected")
            return

        # Put the arguments through the debugger
        try:
            arg_list = pwndbg.dbg.lex_args(argument)
        except (TypeError, pwndbg.dbg_mod.Error):
            pwndbg.exception.handle(self.function.__name__)
            return

        # Put the arguments through argparse
        try:
            kwargs = vars(self.parser.parse_args(arg_list))
        except SystemExit:
            # argparse complained about incorrect usage or printed
            # help and exited. Either way the appropriate message
            # is already printed and we shouldn't call the function.
            return

        try:
            self.repeat = self.check_repeated(argument, from_tty)
            # Call this object, same as `self(**kwargs)` but faster.
            self.__call__(**kwargs)
        finally:
            self.repeat = False

    def check_repeated(self, argument: str, from_tty: bool) -> bool:
        """
        Keep a record of all commands which come from the TTY.

        Returns:
            True if this command was executed by the user just hitting "enter".
        """
        # Don't care unless it's interactive use
        if not from_tty:
            return False

        last_line = pwndbg.dbg.history(1)

        # No history
        if not last_line:
            return False

        number, command = last_line[-1]
        # A new command was entered by the user
        if number not in CommandObj.history:
            CommandObj.history[number] = command
            return False

        # Somehow the command is different than we got before?
        if not command.endswith(argument):
            return False

        return True

    def __call__(self, *args: Any, **kwargs: Any) -> str | None:
        try:
            return self.function(*args, **kwargs)
        except TypeError:
            print(f"{self.command_name}: {self.description}")
            pwndbg.exception.handle(self.function.__name__)
        except ConnectionRefusedError:
            print(message.error("Connection Refused Exception."))
            print(message.hint("Did the decompiler integration connection die?"), end="")
            # If yes, we need to throw the connection out and fix up the manager's
            # state. The manager has not yet realized that the connection is doomed,
            # so we can check like this if we *were* connected.
            if pwndbg.dintegration.manager.is_connected():
                decompiler_name = pwndbg.dintegration.manager.decompiler_name()
                pwndbg.dintegration.manager.disconnect()
                print(message.hint(f" Automatically disabled {decompiler_name} integration."))
                print("Feel free to re-enable manually.")
            else:
                print()
        except TypeNotRecoveredError as e:
            print(message.error(f"recovering {e.name} failed with error:"))
            print(e)
            if "CONFIG_RANDSTRUCT" in pwndbg.aglib.kernel.kconfig():
                print(
                    message.warn(
                        "please note that some structs may not be recoverable when CONFIG_RANDSTRUCT=y"
                    )
                )

        except Exception:
            pwndbg.exception.handle(self.function.__name__)
        return None


class Command:
    """
    Parametrized decorator for functions that serve as pwndbg commands.

    Always use this to decorate your commands.
    """

    def __init__(
        self,
        parser_or_desc: argparse.ArgumentParser | str,
        *,  # All further parameters are not positional
        category: CommandCategory,
        command_name: str | None = None,
        aliases: list[str] = [],
        examples: str = "",
        notes: str = "",
        only_debuggers: set[pwndbg.dbg_mod.DebuggerType] = None,
        exclude_debuggers: set[pwndbg.dbg_mod.DebuggerType] = None,
    ) -> None:
        # Setup an ArgumentParser even if we were only passed a description.
        if isinstance(parser_or_desc, str):
            self.parser = argparse.ArgumentParser(description=parser_or_desc)
        else:
            assert isinstance(parser_or_desc, argparse.ArgumentParser)
            self.parser = parser_or_desc

        self.category = category
        self.command_name = command_name
        self.aliases = aliases
        self.examples = examples
        self.notes = notes
        self.only_debuggers = only_debuggers
        self.exclude_debuggers = exclude_debuggers

    def __call__(self, function: Callable[..., Any]) -> CommandObj:
        # Since this is the __call__ of a parametrized decorator, it is
        # invoked during decoration, and it must return a callable object
        # i.e. the "real" decorator of the function.

        # If this command is not valid for this debugger, do not even
        # pass it to ComandObj to be registered with the debugger API.
        # Also make sure it raises an error if it is called from the code.
        if self.only_debuggers is not None and pwndbg.dbg.name() not in self.only_debuggers:

            def decorator(*args: Any, **kwargs: Any) -> None:
                raise InvalidDebuggerError(
                    f"This command cannot be used in {pwndbg.dbg.name()}.\n"
                    f"It is only valid for {self.only_debuggers}."
                )

            return decorator  # type: ignore[return-value]
        if self.exclude_debuggers is not None and pwndbg.dbg.name() in self.exclude_debuggers:

            def decorator(*args: Any, **kwargs: Any) -> None:
                raise InvalidDebuggerError(
                    f"This command cannot be used in {pwndbg.dbg.name()}.\n"
                    f"It is invalid for {self.exclude_debuggers}."
                )

            return decorator  # type: ignore[return-value]

        # Since CommandObj has __call__ defined, an instance of it is a
        # callable object (which essentially decorates the function).
        return CommandObj(
            function,
            self.parser,
            self.command_name,
            self.category,
            self.aliases,
            self.examples,
            self.notes,
        )


def fix(
    arg: pwndbg.dbg_mod.Value | str, sloppy: bool = False, quiet: bool = True, reraise: bool = False
) -> str | pwndbg.dbg_mod.Value | None:
    """Fix a single command-line argument coming from the CLI.

    Arguments:
        arg: Original string representation (e.g. '0', '$rax', '$rax+44')
        sloppy: If ``arg`` cannot be evaluated, return ``arg``. (default: False)
        quiet: If an error occurs, suppress it. (default: True)
        reraise: If an error occurs, raise the exception. (default: False)

    Returns:
        Ideally a ``Value`` object.  May return a ``str`` if ``sloppy==True``.
        May return ``None`` if ``sloppy == False and reraise == False``.
    """
    if isinstance(arg, pwndbg.dbg_mod.Value):
        return arg

    frame = pwndbg.dbg.selected_frame()
    try:
        target: pwndbg.dbg_mod.Frame | pwndbg.dbg_mod.Process = (
            frame if frame else pwndbg.dbg.selected_inferior()
        )
    except pwndbg.dbg_mod.NoInferior:
        raise AssertionError("Reached command expression evaluation with no frame or inferior")

    # Try to evaluate the expression in the local, or, failing that, global
    # context.
    try:
        return target.evaluate_expression(arg)
    except Exception:
        pass

    ex = None
    try:
        # This will fail if gdblib is not available. While the next check
        # alleviates the need for this call, it's not really equivalent, and
        # we'll need a debugger-agnostic version of regs.fix() if we want to
        # completely get rid of this call. We can't do that now because there's
        # no debugger-agnostic architecture functions. Those will come later.
        #
        # TODO: Port architecutre functions and `pwndbg.gdblib.regs.fix` to debugger-agnostic API and remove this.
        arg = pwndbg.aglib.regs.fix(arg)
        return target.evaluate_expression(arg)
    except Exception as e:
        ex = e

    # If that fails, try to treat the argument as the name of a register, and
    # see if that yields anything.
    if frame:
        regs = frame.regs()
        arg = arg.strip()
        if arg.startswith("$"):
            arg = arg[1:]
        reg = regs.by_name(arg)
        if reg:
            return reg

    # If both fail, check whether we want to print or re-raise the error we
    # might've gotten from `evaluate_expression`.
    if ex:
        if not quiet:
            print(ex)
        if reraise:
            raise ex

    if sloppy:
        return arg

    return None


def fix_reraise(*a: Any, **kw: Any) -> str | pwndbg.dbg_mod.Value | None:
    # Type error likely due to https://github.com/python/mypy/issues/6799
    return fix(*a, reraise=True, **kw)  # type: ignore[misc]


def fix_reraise_arg(arg: Any) -> pwndbg.dbg_mod.Value:
    """fix_reraise wrapper for evaluating command arguments"""
    try:
        # Will always return pwndbg.dbg_mod.Value because
        # sloppy=False (not str) and reraise=True (not None)
        fixed = fix(arg, sloppy=False, quiet=True, reraise=True)
        assert isinstance(fixed, pwndbg.dbg_mod.Value)
        return fixed
    except pwndbg.dbg_mod.Error as dbge:
        raise argparse.ArgumentTypeError(f"debugger couldn't resolve argument '{arg}': {dbge}")


def fix_int(*a: Any, **kw: Any) -> int:
    return int(fix(*a, **kw))


def fix_int_reraise(*a: Any, **kw: Any) -> int:
    return fix_int(*a, reraise=True, **kw)


def fix_int_reraise_arg(arg: Any) -> int:
    """fix_int_reraise wrapper for evaluating command arguments"""
    try:
        fixed: pwndbg.dbg_mod.Value = fix_reraise_arg(arg)
        if fixed.type.code == pwndbg.dbg_mod.TypeCode.FUNC:
            # Fixes issues with function ptrs (e.g. passing in `malloc`).
            func_addr = fixed.address
            if func_addr is None:
                raise argparse.ArgumentTypeError(
                    f"couldn't convert '{arg}' ({fixed.type.name_to_human_readable}) to int: Function is not addressable."
                )
            return int(func_addr)
        return int(fixed)
    except pwndbg.dbg_mod.Error as e:
        raise argparse.ArgumentTypeError(
            f"couldn't convert '{arg}' ({fixed.type.name_to_human_readable}) to int: {e}"
        )


def func_name(function: Callable[P, T]) -> str:
    return function.__name__.replace("_", "-")


def OnlyWhenLocal(function: Callable[P, T]) -> Callable[P, T | None]:
    @functools.wraps(function)
    def _OnlyWhenLocal(*a: P.args, **kw: P.kwargs) -> T | None:
        if not pwndbg.aglib.remote.is_remote():
            return function(*a, **kw)

        msg = f'The "remote" target does not support "{function.__name__}".'

        if pwndbg.dbg.is_gdblib_available():
            msg += ' Try "help target" or "continue".'

        log.error(msg)
        return None

    return _OnlyWhenLocal


def OnlyWithFile(function: Callable[P, T]) -> Callable[P, T | None]:
    @functools.wraps(function)
    def _OnlyWithFile(*a: P.args, **kw: P.kwargs) -> T | None:
        if pwndbg.aglib.proc.exe():
            return function(*a, **kw)
        if pwndbg.aglib.qemu.is_qemu():
            log.error("Could not determine the target binary on QEMU.")
        else:
            log.error(f"{func_name(function)}: There is no file loaded.")
        return None

    return _OnlyWithFile


def OnlyWhenQemuKernel(function: Callable[P, T]) -> Callable[P, T | None]:
    @functools.wraps(function)
    def _OnlyWhenQemuKernel(*a: P.args, **kw: P.kwargs) -> T | None:
        if pwndbg.aglib.qemu.is_qemu_kernel():
            return function(*a, **kw)
        log.error(
            f"{func_name(function)}: This command may only be run when debugging the Linux kernel in QEMU."
        )
        return None

    return _OnlyWhenQemuKernel


def OnlyWhenUserspace(function: Callable[P, T]) -> Callable[P, T | None]:
    @functools.wraps(function)
    def _OnlyWhenUserspace(*a: P.args, **kw: P.kwargs) -> T | None:
        if not pwndbg.aglib.qemu.is_qemu_kernel():
            return function(*a, **kw)
        log.error(
            f"{func_name(function)}: This command may only be run when not debugging a QEMU kernel target."
        )
        return None

    return _OnlyWhenUserspace


def OnlyWithKernelDebugInfo(function: Callable[P, T]) -> Callable[P, T | None]:
    @functools.wraps(function)
    def _OnlyWithKernelDebugInfo(*a: P.args, **kw: P.kwargs) -> T | None:
        if pwndbg.aglib.kernel.has_debug_info():
            return function(*a, **kw)
        log.error(
            f"{func_name(function)}: This command may only be run when debugging a Linux kernel with debug info."
        )
        return None

    return _OnlyWithKernelDebugInfo


def OnlyWithKernelSymbols(function: Callable[P, T]) -> Callable[P, T | None]:
    @functools.wraps(function)
    def _OnlyWithKernelSymbols(*a: P.args, **kw: P.kwargs) -> T | None:
        if pwndbg.aglib.kernel.has_debug_symbols():
            return function(*a, **kw)
        log.error(
            f"{func_name(function)}: This command may only be run when debugging a Linux kernel with symbols.\n"
            + message.hint(
                "Check out vmlinux-to-elf to get them easily (https://github.com/marin-m/vmlinux-to-elf) or compile the kernel yourself."
            )
        )
        return None

    return _OnlyWithKernelSymbols


def OnlyWhenPagingEnabled(function: Callable[P, T]) -> Callable[P, T | None]:
    @functools.wraps(function)
    def _OnlyWhenPagingEnabled(*a: P.args, **kw: P.kwargs) -> T | None:
        if pwndbg.aglib.kernel.paging_enabled():
            return function(*a, **kw)
        log.error(f"{func_name(function)}: This command may only be run when paging is enabled.")
        return None

    return _OnlyWhenPagingEnabled


def WarnOnKernelConfigRandstruct(function: Callable[P, T]) -> Callable[P, T | None]:
    @functools.wraps(function)
    def _WarnOnKernelConfigRandstruct(*a: P.args, **kw: P.kwargs) -> T | None:
        if (
            not pwndbg.aglib.kernel.has_debug_info()
            and "CONFIG_RANDSTRUCT" in pwndbg.aglib.kernel.kconfig()
        ):
            log.warning("command output may be inaccurate because CONFIG_RANDSTRUCT=y")
        return function(*a, **kw)

    return _WarnOnKernelConfigRandstruct


def OnlyWhenRunning(
    func_when_no_kwargs: Callable[P, T] | None = None, *, allow_core: bool = True
) -> Callable[[Callable[P, T]], Callable[P, T | None]] | Callable[P, T | None]:
    def decorator(func: Callable[P, T]) -> Callable[P, T | None]:
        @functools.wraps(func)
        def _OnlyWhenRunning(*a: P.args, **kw: P.kwargs) -> T | None:
            if pwndbg.aglib.proc.alive() and not (
                not allow_core and pwndbg.aglib.proc.is_core_file()
            ):
                return func(*a, **kw)
            log.error(f"{func_name(func)}: The program is not being run.")
            return None

        return _OnlyWhenRunning

    if func_when_no_kwargs is None:
        return decorator
    return decorator(func_when_no_kwargs)


def OnlyWithTcache(function: Callable[P, T]) -> Callable[P, T | None]:
    @functools.wraps(function)
    def _OnlyWithTcache(*a: P.args, **kw: P.kwargs) -> T | None:
        assert isinstance(pwndbg.aglib.heap.current, GlibcMemoryAllocator)
        if pwndbg.aglib.heap.current.has_tcache():
            return function(*a, **kw)
        log.error(
            f"{func_name(function)}: This version of GLIBC was not compiled with tcache support."
        )
        return None

    return _OnlyWithTcache


def OnlyWhenHeapIsInitialized(function: Callable[P, T]) -> Callable[P, T | None]:
    @functools.wraps(function)
    def _OnlyWhenHeapIsInitialized(*a: P.args, **kw: P.kwargs) -> T | None:
        if pwndbg.aglib.heap.current is not None and pwndbg.aglib.heap.current.is_initialized():
            return function(*a, **kw)
        log.error(f"{func_name(function)}: Heap is not initialized yet.")
        return None

    return _OnlyWhenHeapIsInitialized


def _try2run_heap_command(function: Callable[P, T], *a: P.args, **kw: P.kwargs) -> T | None:
    e = log.error
    w = log.warning
    # Note: We will still raise the error for developers when exception-* is set to "on"
    try:
        return function(*a, **kw)
    except SymbolNotRecoveredError as err:
        e(f"{func_name(function)}: Fail to resolve the symbol: `{err.name}`")
        if "thread_arena" == err.name:
            w(
                "You are probably debugging a multi-threaded target without debug symbols, so we failed to determine which arena is used by the current thread.\n"
                "To resolve this issue, you can use the `arenas` command to list all arenas, and use `set thread-arena <addr>` to set the current thread's arena address you think is correct.\n"
            )
        else:
            w(
                f"You can try to determine the libc symbols addresses manually and set them appropriately. For this, see the `heap-config` command output and set the config for `{err.name}`."
            )
        if pwndbg.config.exception_verbose or pwndbg.config.exception_debugger:
            raise err

        pwndbg.exception.inform_verbose_and_debug()
    except Exception as err:
        e(f"{func_name(function)}: An unknown error occurred when running this command.")
        if isinstance(pwndbg.aglib.heap.current, HeuristicHeap):
            w(
                "Maybe you can try to determine the libc symbols addresses manually, set them appropriately and re-run this command. For this, see the `heap-config` command output and set the `main_arena`, `mp_`, `global_max_fast`, `tcache` and `thread_arena` addresses."
            )
        else:
            w("You can try `set resolve-heap-via-heuristic force` and re-run this command.\n")
        if pwndbg.config.exception_verbose or pwndbg.config.exception_debugger:
            raise err

        pwndbg.exception.inform_verbose_and_debug()
    return None


def OnlyWithResolvedHeapSyms(function: Callable[P, T]) -> Callable[P, T | None]:
    @functools.wraps(function)
    def _OnlyWithResolvedHeapSyms(*a: P.args, **kw: P.kwargs) -> T | None:
        e = log.error
        w = log.warning

        # Operating under the assumption that the pwndbg/libc/ code can figure out
        # that we are using glibc with at least as good accuracy as the ptmalloc code.
        if pwndbg.libc.which() != pwndbg.libc.LibcType.GLIBC:
            e(f"The currently active libc isn't glibc. It's {pwndbg.libc.which().value}.")
            return None

        if (
            isinstance(pwndbg.aglib.heap.current, HeuristicHeap)
            and pwndbg.config.resolve_heap_via_heuristic == "auto"
            and DebugSymsHeap().can_be_resolved()
        ):
            # In auto mode, we will try to use the debug symbols if possible
            pwndbg.aglib.heap.current = DebugSymsHeap()

        if (
            pwndbg.aglib.heap.current is not None
            and isinstance(pwndbg.aglib.heap.current, GlibcMemoryAllocator)
            and pwndbg.aglib.heap.current.can_be_resolved()
        ):
            return _try2run_heap_command(function, *a, **kw)

        static = not pwndbg.dbg.selected_inferior().is_dynamically_linked()
        if (
            isinstance(pwndbg.aglib.heap.current, DebugSymsHeap)
            and pwndbg.config.resolve_heap_via_heuristic == "auto"
        ):
            # In auto mode, if the debug symbols are not enough, we will try to use the heuristic if possible
            heuristic_heap = HeuristicHeap()
            if heuristic_heap.can_be_resolved():
                pwndbg.aglib.heap.current = heuristic_heap
                w(
                    "pwndbg will try to resolve the heap symbols via heuristic now since we cannot resolve the heap via the debug symbols.\n"
                    "This might not work in all cases. Use `help set resolve-heap-via-heuristic` for more details.\n"
                )
                return _try2run_heap_command(function, *a, **kw)
            if static:
                e(
                    "Can't find GLIBC version required for this command to work since this is a statically linked binary"
                )
                w(
                    "Please set the GLIBC version you think the target binary was compiled (using `set glibc <version>` command; e.g. 2.32) and re-run this command."
                )
            else:
                e(
                    "Can't find GLIBC version required for this command to work, maybe is because GLIBC is not loaded yet."
                )
                w(
                    "If you believe the GLIBC is loaded or this is a statically linked binary. "
                    "Please set the GLIBC version you think the target binary was compiled (using `set glibc <version>` command; e.g. 2.32) and re-run this command"
                )
        elif (
            isinstance(pwndbg.aglib.heap.current, DebugSymsHeap)
            and pwndbg.config.resolve_heap_via_heuristic == "force"
        ):
            e(
                "You are forcing to resolve the heap symbols via heuristic, but we cannot resolve the heap via the debug symbols."
            )
            w("Use `set resolve-heap-via-heuristic auto` and re-run this command.")
        else:
            # Note: Should not see this error, but just in case
            e("An unknown error occurred when resolved the heap.")
            pwndbg.exception.inform_report_issue("An unknown error occurred when resolved the heap")
        return None

    return _OnlyWithResolvedHeapSyms


def sloppy_gdb_parse(s: str) -> int | str:
    """
    This function should be used as ``argparse.ArgumentParser`` .add_argument method's `type` helper.

    This makes the type being parsed as gdb value and if that parsing fails,
    a string is returned.

    :param s: String.
    :return: Whatever gdb.parse_and_eval returns or string.
    """

    frame = pwndbg.dbg.selected_frame()
    try:
        target: pwndbg.dbg_mod.Frame | pwndbg.dbg_mod.Process = (
            frame if frame else pwndbg.dbg.selected_inferior()
        )
    except pwndbg.dbg_mod.NoInferior:
        raise AssertionError("Reached command expression evaluation with no frame or inferior")

    try:
        val = pwndbg.aglib.symbol.lookup_symbol(s) or target.evaluate_expression(s)
        if val.type.code == pwndbg.dbg_mod.TypeCode.FUNC:
            return int(val.address)
        return int(val)
    except (TypeError, pwndbg.dbg_mod.Error):
        return s


def AddressExpr(s: str) -> int:
    """
    Parses an address expression. Returns an int.
    """
    val = sloppy_gdb_parse(s)

    if not isinstance(val, int):
        raise argparse.ArgumentTypeError(f"Incorrect address (or GDB expression): {s}")

    return val


def HexOrAddressExpr(s: str) -> int:
    """
    Parses string as hexadecimal int or an address expression. Returns an int.
    (e.g. '1234' will return 0x1234)
    """
    try:
        return int(s, 16)
    except ValueError:
        return AddressExpr(s)


def load_commands() -> None:
    # pylint: disable=import-outside-toplevel
    import pwndbg.dbg_mod

    if pwndbg.dbg.is_gdblib_available():
        import pwndbg.commands.ai
        import pwndbg.commands.attachp
        import pwndbg.commands.branch
        import pwndbg.commands.cymbol
        import pwndbg.commands.got_tracking
        import pwndbg.commands.ignore
        import pwndbg.commands.ipython_interactive
        import pwndbg.commands.killthreads
        import pwndbg.commands.peda
        import pwndbg.commands.ptmalloc2_tracking
        import pwndbg.commands.reload
        import pwndbg.commands.ropper
        import pwndbg.commands.segments
        import pwndbg.commands.updown

    import pwndbg.commands.argv
    import pwndbg.commands.aslr
    import pwndbg.commands.asm
    import pwndbg.commands.auxv
    import pwndbg.commands.binder
    import pwndbg.commands.buddydump
    import pwndbg.commands.canary
    import pwndbg.commands.checksec
    import pwndbg.commands.comments
    import pwndbg.commands.commpage
    import pwndbg.commands.config
    import pwndbg.commands.context
    import pwndbg.commands.cpsr
    import pwndbg.commands.cyclic
    import pwndbg.commands.decompiler_integration
    import pwndbg.commands.dev
    import pwndbg.commands.distance
    import pwndbg.commands.dt
    import pwndbg.commands.dumpargs
    import pwndbg.commands.elf
    import pwndbg.commands.errno
    import pwndbg.commands.flags
    import pwndbg.commands.gdt
    import pwndbg.commands.godbg
    import pwndbg.commands.got
    import pwndbg.commands.hex2ptr
    import pwndbg.commands.hexdump
    import pwndbg.commands.hijack_fd
    import pwndbg.commands.jemalloc
    import pwndbg.commands.kbase
    import pwndbg.commands.kbpf
    import pwndbg.commands.kchecksec
    import pwndbg.commands.kcmdline
    import pwndbg.commands.kconfig
    import pwndbg.commands.kcurrent
    import pwndbg.commands.kdmabuf
    import pwndbg.commands.kdmesg
    import pwndbg.commands.klookup
    import pwndbg.commands.kmem_trace
    import pwndbg.commands.kmod
    import pwndbg.commands.knft
    import pwndbg.commands.ksyscalls
    import pwndbg.commands.ktask
    import pwndbg.commands.kversion
    import pwndbg.commands.leakfind
    import pwndbg.commands.libcinfo
    import pwndbg.commands.linkmap
    import pwndbg.commands.mallocng
    import pwndbg.commands.memoize
    import pwndbg.commands.mmap
    import pwndbg.commands.mprotect
    import pwndbg.commands.msr
    import pwndbg.commands.nearpc
    import pwndbg.commands.next
    import pwndbg.commands.onegadget
    import pwndbg.commands.p2p
    import pwndbg.commands.paging
    import pwndbg.commands.parse_seccomp
    import pwndbg.commands.patch
    import pwndbg.commands.pie
    import pwndbg.commands.plist
    import pwndbg.commands.probeleak
    import pwndbg.commands.procinfo
    import pwndbg.commands.profiler
    import pwndbg.commands.ptmalloc2
    import pwndbg.commands.pwndbg_
    import pwndbg.commands.radare2
    import pwndbg.commands.retaddr
    import pwndbg.commands.rizin
    import pwndbg.commands.rop
    import pwndbg.commands.saved_register_frames
    import pwndbg.commands.search
    import pwndbg.commands.sigreturn
    import pwndbg.commands.slab
    import pwndbg.commands.spray
    import pwndbg.commands.start
    import pwndbg.commands.strings
    import pwndbg.commands.telescope
    import pwndbg.commands.tips
    import pwndbg.commands.tls
    import pwndbg.commands.valist
    import pwndbg.commands.version
    import pwndbg.commands.vmmap
    import pwndbg.commands.windbg
    import pwndbg.commands.xinfo
    import pwndbg.commands.xor
